package edu.bu.signstream.common.db.ss3;

import edu.bu.signstream.common.db.XMLWriter;
import edu.bu.signstream.common.util.UserInterfaceUtil;
import edu.bu.signstream.common.util.Util;
import edu.bu.signstream.common.util.vo.Constants;
import edu.bu.signstream.common.util.vo.ss3.Reference;
import edu.bu.signstream.common.util.vo.ss3.codingScheme.SS3CodingScheme;
import edu.bu.signstream.common.util.vo.ss3.codingScheme.SS3Field;
import edu.bu.signstream.common.util.vo.ss3.codingScheme.SS3FieldValue;
import edu.bu.signstream.common.util.vo.ss3.db.*;
import edu.bu.signstream.common.util.vo.ss3.media.SS3MediaLibrary;
import edu.bu.signstream.common.util.vo.ss3.note.SS3Note;
import edu.bu.signstream.common.util.vo.ss3.participants.SS3Participant;
import edu.bu.signstream.grepresentation.fields.gestureField.GestureEntityProperties;
import edu.bu.signstream.grepresentation.fields.nonManualField.GlossEntityProperties;
import edu.bu.signstream.grepresentation.reguestsHandler.SS3Singleton;
import edu.bu.signstream.time.Rounding;
import edu.bu.signstream.time.Time;
import edu.bu.signstream.time.TimeUnit;
import edu.bu.signstream.ui.SS3SignStreamApplication;
import edu.bu.signstream.ui.SerchableDirectory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import javax.swing.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.logging.Level;
import java.util.stream.Collectors;

import static edu.bu.signstream.common.db.ss4.SS4Util.*;

/**
 * Temporary class to save SS3 database.
 */
public class SS3Writer {

    private void addStartEndTimeAttributes(Element el, Time start, Time end) {
    	writeLong(el, "S", start.value());
    	writeLong(el, "E", end.value());
    	// collections saved since 3.4.0 are in frames
    }

	private Element createUtilsElement(SS3Database db, Document document) {
		Element root = document.createElement("UTILS");
		root.setAttribute("TP_PIXEL_IN_TIME_FRAME",
				"" + db.getTemporalPartitionPixelInTimeFrame());
		root.setAttribute("TP_ONSCREEN_LOCATION",
				"" + db.getTemporalPartitionOnScreenLocation());
		root.setAttribute("TP_MOVIE_TIME",
				"" + db.getTemporalPartitionMovieTime());
		
		return root;
	}

    private Element createDistributorElement(SS3DBProfile prf, Document document) {
        Element root = document.createElement("DISTRIBUTOR");
        root.appendChild(document.createTextNode(prf.getDistributor()));
        return root;
    }

    private Element createReferencesElement(SS3Database db, Document document) {
		Element root = document.createElement("REFERENCES");
		root.appendChild(createCodingSchemasElement(db, document));
		root.appendChild(createParticipantsElement(db, document));
		root.appendChild(createAuthorsElement(db, document));
		return root;
	}

	private Element createCodingSchemasElement(SS3Database db, Document document) {
		Element root = document.createElement("CODING-SCHEMES");
		ArrayList csRef = db.getCodingSchemeRef();
		for (int i = 0; i < csRef.size(); i++) {
			Reference ref = (Reference) csRef.get(i);
			Element rEl = document.createElement("REF");
			rEl.setAttribute("ID", "" + ref.getId());
			rEl.setAttribute("FILE_NAME", "" + ref.getFileName());
			rEl.setAttribute("SIGNATURE", "" + ref.getSignature());
			root.appendChild(rEl);
		}
		return root;
	}

	private Element createParticipantsElement(SS3Database db, Document document) {
		Element root = document.createElement("PARTICIPANTS");
		ArrayList<SS3Participant> participants = db.getSelectedParticipants();

		for (SS3Participant participant : participants) {
			Element pEl = document.createElement("REF");
			pEl.setAttribute("ID", "" + participant.getId());
			pEl.setAttribute("LABEL", "" + participant.getLabel());
			pEl.setAttribute("BACKGROUND", "" + participant.getBackground());
			pEl.setAttribute("COMMENTS", "" + participant.getComments());
			pEl.setAttribute("DOMINANT_HAND", "" + participant.getDominantHand());
			pEl.setAttribute("FULL_NAME", "" + participant.getFullName());
			pEl.setAttribute("GENDER", "" + participant.getGender());
			pEl.setAttribute("NATIVE_LANGUAGE", "" + participant.getNativeLanguage());
			pEl.setAttribute("PARENT_INFO", "" + participant.getParentsInfo());
			pEl.setAttribute("YEAR_OF_BIRTH", "" + participant.getYearOfBirth());

			if (participant.getGraphFileID() != null && !participant.getGraphFileID().isEmpty()) {
				pEl.setAttribute("GRAPH_FILE_ID", participant.getGraphFileID());
			}
			root.appendChild(pEl);
		}
		return root;
	}

	private Element createAuthorsElement(SS3Database db, Document document) {
		Element root = document.createElement("ANNOTATORS");
		ArrayList<SS3Participant> annotators = db.getSelectedAnnotators();

		for (SS3Participant annotator : annotators) {
			Element aEl = document.createElement("REF");
			aEl.setAttribute("ID", "" + annotator.getId());
			aEl.setAttribute("LABEL", "" + annotator.getLabel());
			root.appendChild(aEl);
		}
		return root;
	}
	
	protected static Element createCodingSchemeElement(SS3CodingScheme localScheme, Document document) {
		Element root = document.createElement("LOCAL-CODING-SCHEME");
		
		SS3CodingScheme defScheme = SS3Singleton.getSS3CodingSchemeCollection().getDefaultCodingScheme();

		if (localScheme == null)
			return root;
		
		Element fieldsEl = document.createElement("FIELDS");

		HashMap localFields = localScheme.getFields();

		Set keys = localFields.keySet();
		Iterator it = keys.iterator();
		while (it.hasNext()) {
			Object obj = localFields.get(it.next());
			SS3Field localField = (SS3Field) obj;
			
			SS3Field defaultField = defScheme.getField(localField.getId());

			if (localField.equals(defaultField))
			    continue;

			String name = localField.getName() == null ? "" : localField.getName();
			String label = localField.getLabel() == null ? "" : localField.getLabel();
			String prefix = localField.getPrefix() == null ? "" : localField.getPrefix();
			String color = localField.getColor() == null ? "" : ""
					+ Util.color2HexString(localField.getColor());
			String type = localField.getType() == null ? "" : localField.getType();
			String alignment = localField.getAlignment() == null ? "" : localField.getAlignment().toString();
			String group = localField.getGroup() == null ? "" : localField.getGroup();
			String order = localField.getOrder() == -1 ? "" : String.valueOf(localField.getOrder());
			boolean deleted = localField.isDeleted();

			boolean appendFieldEl = false;
			Element fieldEl = document.createElement("FIELD");
			fieldEl.setAttribute("ID", localField.getId());
			
			boolean appendFieldAttrsEl = false;
			Element fieldAttrsEl = document.createElement("FIELD-ATTRIBUTES");
			
			if ( ! name.isEmpty() && (defaultField == null || ! name.equals(defaultField.getName()))) {
				Element fieldAttrEl = document.createElement("FIELD-ATTRIBUTE");
				fieldAttrEl.setAttribute("NAME", name);
				fieldAttrsEl.appendChild(fieldAttrEl);
				
				appendFieldEl = true;
				appendFieldAttrsEl = true;
			}
			if ( ! label.isEmpty() && (defaultField == null || ! label.equals(defaultField.getLabel()))) {
				Element fldAttrEl = document.createElement("FIELD-ATTRIBUTE");
				fldAttrEl.setAttribute("LABEL", label);
				fieldAttrsEl.appendChild(fldAttrEl);
				
				appendFieldEl = true;
				appendFieldAttrsEl = true;
			}
			if ( ! prefix.isEmpty() && (defaultField == null || ! prefix.equals(defaultField.getPrefix()))) {
				Element fldAttrEl = document.createElement("FIELD-ATTRIBUTE");
				fldAttrEl.setAttribute("PREFIX", prefix);
				fieldAttrsEl.appendChild(fldAttrEl);
				
				appendFieldEl = true;
				appendFieldAttrsEl = true;
			}
			if ( ! color.isEmpty()) {
				boolean writeIt = false;
				
				if (defaultField != null) {
					var defaultFieldColor = defaultField.getColor();
					if (defaultFieldColor != null) {
						var defaultFieldHexColor = Util.color2HexString(defaultFieldColor);
						if ( ! color.equals(defaultFieldHexColor))
							writeIt = true;
					}
				} else {
					if (! color.equals(Util.color2HexString(SS3Field.defaultColor)))
						writeIt = true;
				}
				
				if (writeIt) {
					Element fldAttrEl = document.createElement("FIELD-ATTRIBUTE");
					fldAttrEl.setAttribute("COLOR", color);
					fieldAttrsEl.appendChild(fldAttrEl);
					
					appendFieldEl = true;
					appendFieldAttrsEl = true;
				}
			}
			if ( ! type.isEmpty() && (defaultField == null || ! type.equals(defaultField.getType()))) {
				Element fldAttrEl = document.createElement("FIELD-ATTRIBUTE");
				fldAttrEl.setAttribute("TYPE", type);
				fieldAttrsEl.appendChild(fldAttrEl);
				
				appendFieldEl = true;
				appendFieldAttrsEl = true;
			}
			if ( ! alignment.isEmpty() && (defaultField == null || ! alignment.equals(defaultField.getAlignment().toString()))) {
				Element fldAttrEl = document.createElement("FIELD-ATTRIBUTE");
				fldAttrEl.setAttribute("ALIGNMENT", alignment);
				fieldAttrsEl.appendChild(fldAttrEl);
				
				appendFieldEl = true;
				appendFieldAttrsEl = true;
			}
			if ( ! group.isEmpty() && (defaultField == null || ! group.equals(defaultField.getGroup()))) {
				Element fldAttrEl = document.createElement("FIELD-ATTRIBUTE");
				fldAttrEl.setAttribute("GROUP", group);
				fieldAttrsEl.appendChild(fldAttrEl);
				
				appendFieldEl = true;
				appendFieldAttrsEl = true;
			}
			if ( !order.isEmpty() && (defaultField == null || ! order.equals(String.valueOf(defaultField.getOrder())))) {
				Element fldAttrEl = document.createElement("FIELD-ATTRIBUTE");
				fldAttrEl.setAttribute("ORDER", order);
				fieldAttrsEl.appendChild(fldAttrEl);
				
				appendFieldEl = true;
				appendFieldAttrsEl = true;
			}
			if (defaultField != null && deleted != defaultField.isDeleted()) {
				fieldEl.setAttribute("DELETED", "T");
				
				appendFieldEl = true;
			}
			
			if (appendFieldAttrsEl) {
				fieldEl.appendChild(fieldAttrsEl);
			}
			
			boolean appendFieldValuesEl = false;
			Element fieldValuesEl = document.createElement("FIELD-VALUES");
			
			ArrayList<SS3FieldValue> localFieldValues = localField.getValuesAsList();
			ArrayList<SS3FieldValue> defaultFieldValues = null;
			ArrayList<SS3FieldValue> deletedFieldValues = new ArrayList<SS3FieldValue>();
			
			if (defaultField != null) {
				defaultFieldValues = defaultField.getValuesAsList();
				
				defaultFieldValues.stream()
						.filter(val -> localField.getFieldValue(val.getId()) == null)
						.forEach(val -> deletedFieldValues.add(val));
			}
			
			for (SS3FieldValue localFieldValue : localFieldValues) {
				SS3FieldValue defaultFieldValue = null;
				if (defaultField != null)
					defaultFieldValue = defaultField.getFieldValue(localFieldValue.getId());
				
				if (localFieldValue.equals(defaultFieldValue))
					continue;

				String vName = localFieldValue.getName() == null ? "" : localFieldValue.getName();
				String vLabel = localFieldValue.getLabel() == null ? "" : localFieldValue.getLabel();
				String vType = localFieldValue.getType() == null ? "" : localFieldValue.getType();
				String iid = localFieldValue.getIid() == null ? "" : localFieldValue.getIid();

				boolean appendFieldValueEl = false;
				Element fieldValueEl = document.createElement("FIELD-VALUE");
				fieldValueEl.setAttribute("ID", localFieldValue.getId());

				boolean appendFieldValueAttrsEl = false;
				Element fieldValueAttrsEl = document.createElement("FIELD-VALUE-ATTRIBUTES");
				
				if ( ! vName.isEmpty() && (defaultFieldValue == null || ! vName.equals(defaultFieldValue.getName()))) {
					Element fldValAttrEl = document.createElement("FIELD-VALUE-ATTRIBUTE");
					fldValAttrEl.setAttribute("NAME", vName);
					fieldValueAttrsEl.appendChild(fldValAttrEl);
					
					appendFieldValueEl = true;
					appendFieldValueAttrsEl = true;
				}
				if ( ! vLabel.isEmpty() && (defaultFieldValue == null || ! vLabel.equals(defaultFieldValue.getLabel()))) {
					Element fldValAttrEl = document.createElement("FIELD-VALUE-ATTRIBUTE");
					fldValAttrEl.setAttribute("LABEL", vLabel);
					fieldValueAttrsEl.appendChild(fldValAttrEl);
					
					appendFieldValueEl = true;
					appendFieldValueAttrsEl = true;
				}
				if ( ! vType.isEmpty() && (defaultFieldValue == null || ! vType.equals(defaultFieldValue.getType()))) {
					Element fldValAttrEl = document.createElement("FIELD-VALUE-ATTRIBUTE");
					fldValAttrEl.setAttribute("TYPE", vType);
					fieldValueAttrsEl.appendChild(fldValAttrEl);
					
					appendFieldValueEl = true;
					appendFieldValueAttrsEl = true;
				}
				if ( ! iid.isEmpty() && (defaultFieldValue == null || ! iid.equals(defaultFieldValue.getIid()))) {
					Element fieldValueAttrEl = document.createElement("FIELD-VALUE-ATTRIBUTE");
					fieldValueAttrEl.setAttribute("IID", iid);
					fieldValueAttrsEl.appendChild(fieldValueAttrEl);
					
					appendFieldValueEl = true;
					appendFieldValueAttrsEl = true;
				}
				
				if (appendFieldValueAttrsEl) {
					fieldValueEl.appendChild(fieldValueAttrsEl);
					
					appendFieldValuesEl = true;
					appendFieldEl = true;
				}
				
				if (appendFieldValueEl) {
					fieldValuesEl.appendChild(fieldValueEl);
					
					appendFieldValuesEl = true;
					appendFieldEl = true;
				}
			}
			
			for (SS3FieldValue deletedFieldValue: deletedFieldValues) {
				Element fieldValueEl = document.createElement("FIELD-VALUE");
				fieldValueEl.setAttribute("ID", deletedFieldValue.getId());
				fieldValueEl.setAttribute("DELETED", "T");
				
				// below is as if appendFieldValueEl == true
				fieldValuesEl.appendChild(fieldValueEl);
				
				appendFieldValuesEl = true;
				appendFieldEl = true;
			}
			
			if (appendFieldValuesEl) {
				fieldEl.appendChild(fieldValuesEl);
			}
			
			if (appendFieldEl) {
				fieldsEl.appendChild(fieldEl);
			}
		}
		
		root.appendChild(fieldsEl);
		
		return root;
	}

	private Element createCodingSchemeElement(SS3Database db, Document document) {
		SS3CodingScheme localScheme = db.getLocalSS3CodingScheme();
		
		return createCodingSchemeElement(localScheme, document);
	}

	private Element createMediaFilesOnScreenLocationElement(SS3Element element,
			Document document) {
		Element root = document.createElement("MEDIA_FILES_ON_SCREEN_LOCATION");
		HashMap<String, Point> locations = element.getVideoOnScreenLocations();
		Set<String> keySet = locations.keySet();
		Iterator<String> keyIter = keySet.iterator();

		while (keyIter.hasNext()) {
			String key = keyIter.next();
			Point point = locations.get(key);
			Element mlEl = document.createElement("MEDIA-FILE");
			mlEl.setAttribute("FILE_NAME", key);
			mlEl.setAttribute("LOCATION_X", "" + point.x);
			mlEl.setAttribute("LOCATION_Y", "" + point.y);
			root.appendChild(mlEl);
		}

		return root;
	}

    private Element createVideosElement(SS3Database db, Document doc) {
        Element root = doc.createElement("MEDIA-FILES");

        var converter = SS3Singleton.getVideoControlManager()
            .getZoomer()
            .getCoordinateTimeConvertor();

        var duration = db.getVideos().get(0).duration().value();
        var timebase = Math.round(converter.timebase());
        var timestep = Math.round(converter.timestep());

        root.setAttribute("MOVIE_DURATION", "" + duration);
        root.setAttribute("MOVIE_TIMESCALE", "" + timebase);
        root.setAttribute("PACKET_DURATION", "" + timestep);
        
        var videosList = db.getVideos().stream().sorted(
        		Comparator.comparing(n -> n.order())).collect(Collectors.toList());

        for (var v : videosList) {
            Element el = doc.createElement("MEDIA-FILE");

            el.setAttribute("ID", Util.getUniqueNumber());
            el.setAttribute("PATH", v.uri().getPath());
            el.setAttribute("ORDER", String.valueOf(v.order()));

            root.appendChild(el);
        }

        return root;
    }

    private Element createGraphsElement(SS3Database db, Document doc) {
	    Element root = doc.createElement("GRAPH-FILES");

	    for (var g : db.getGraphs()) {
	        Element el = doc.createElement("GRAPH-FILE");
	        el.setAttribute("ID", g.name());
            el.setAttribute("NAME", g.name());
	        el.setAttribute("PATH", g.uri().getPath());

	        root.appendChild(el);
        }

	    return root;
    }

	private Element createGlossWindowElement(SS3Database db, Document document) {
		Element root = document.createElement("WINDOW");
		Element dtEl = document.createElement("DISPLAY-TIME-PERIOD");
		root.appendChild(dtEl);

		SS3GlossWindow window = db.getSS3GlossWindow();
		ArrayList<SS3GlossWindowItem> items = window.getDisplayTimePeriod()
				.getSS3GlossWindowItems();

        writeRefs(root, "VISIBLE", "FIELD", window.getPermanentFieldIDs().stream());

		for (SS3GlossWindowItem item : items) {
			item.synchronize();

			Element sEl = document.createElement("SEGMENT-TIER-UTTERANCE");
			sEl.setAttribute("ID", item.getId());
			sEl.setAttribute("LABEL", item.getLabel());

			addStartEndTimeAttributes(sEl, item.getStartTimeInfo(), item.getEndTimeInfo());

			dtEl.appendChild(sEl);
		}

		Element tpEl = document.createElement("TEMPORAL-PARTITIONS");
		root.appendChild(tpEl);

		ArrayList<SS3TemporalPartition> temporalPartitions = window
				.getTemporalPartitions();

		for (SS3TemporalPartition temporalPartition : temporalPartitions) {
			Element tp1El = document.createElement("TEMPORAL-PARTITION");
			tp1El.setAttribute("ID", temporalPartition.getID());
			tp1El.setAttribute("LABEL", temporalPartition.getLabel());
			tpEl.appendChild(tp1El);

			ArrayList<SS3GlossWindowSegment> segments = temporalPartition
					.getTemporalPartitionSegments();
			for (SS3GlossWindowSegment segment : segments) {
				Element sEl = document
						.createElement("SEGMENT-TIER");
				sEl.setAttribute("ID", segment.getID());
				sEl.setAttribute("LABEL", segment.getLabel());
				String isdefault = segment.isDefault() ? "T" : "F";
				sEl.setAttribute("IS_DEFAULT", "" + isdefault);
				sEl.setAttribute("INDEX_FIELD_ID", segment.getIndexFieldID());
				String participantID = segment.getParticipant() != null ? segment.getParticipant().getId() : "";
				sEl.setAttribute("PARTICIPANT_ID", "" + participantID);

                writeRefs(sEl, "VISIBLE", "FIELD", segment.getPermanentFieldIDs().stream());

                writeRefs(sEl, "VISIBLE", PANEL_TAG, segment.getVisiblePanelIDs().stream());

				SS3GlossWindowItemNote note = segment.getNote();
				if (note != null) {
					String noteStr = note.getNote();
					if (noteStr != null && !noteStr.isEmpty()) {
						sEl.setAttribute("NOTE", noteStr);
					}
				}

				// statement fields exist here, under the segment tier object
				ArrayList<SS3StatementField> statementFields1 = segment
						.getStatementFields();
				for (SS3StatementField statementField : statementFields1) {
					// STATEMENT-FIELD
					Element stmtEl = document
							.createElement("STATEMENT-FIELD");
					stmtEl.setAttribute("FIELD-ID", statementField.getId());

					Color fieldColor = statementField.getColor();
					if (fieldColor != null) {
						stmtEl.setAttribute("COLOR",
								"" + fieldColor.getRGB());
					}

					ArrayList<String> entityIDs = new ArrayList<String>();

					ArrayList<SS3Entity> ss3Entities = statementField.getSS3Entities();
					for (SS3Entity ss3Entity: ss3Entities) {
						entityIDs.add(ss3Entity.getID());
					}

					StringBuffer bf = new StringBuffer();

					for (int counter = 0; counter < entityIDs.size(); counter++) {
						bf.append(entityIDs.get(counter));
						if (counter < entityIDs.size() - 1)
							bf.append(", ");
					}

					stmtEl.setAttribute("ENTITY-IDS", bf.toString());
					sEl.appendChild(stmtEl);
				}

				tp1El.appendChild(sEl);

				ArrayList<SS3GlossWindowItem> segmentItems = segment
						.getSS3GlossWindowItems();
				for (SS3GlossWindowItem segmentItem : segmentItems) {
					segmentItem.synchronize();

					Element wEl = document.createElement("SEGMENT-TIER-UTTERANCE");
					wEl.setAttribute("ID", segmentItem.getId());
					wEl.setAttribute("LABEL", segmentItem.getLabel());

                    writeRefs(wEl, "VISIBLE", "FIELD", segmentItem.getPermanentVisibleFieldIDs().stream());

                    addStartEndTimeAttributes(wEl, segmentItem.getStartTimeInfo(), segmentItem.getEndTimeInfo());

					sEl.appendChild(wEl);

					// 2021-03-23: no longer saving/loading entities from utterances anymore - they are only linked by
					// the parentEntityID relationship - all entites are saved under a segment tier
					// certain fields (e.g. custom fields, only, as of this writing) allow outside utterance bounds
					
//					ArrayList<SS3StatementField> statementFields = segmentItem
//							.getStatementFields();
//					for (SS3StatementField statementField : statementFields) {
//						// STATEMENT-FIELD
//						Element stmtEl = document
//								.createElement("STATEMENT-FIELD");
//						stmtEl.setAttribute("FIELD-ID", statementField.getId());
//
//						Color fieldColor = statementField.getColor();
//						if (fieldColor != null) {
//							stmtEl.setAttribute("COLOR",
//									"" + fieldColor.getRGB());
//						}
//
//						var timeRange = statementField.getDefaultTimeRange();
//						if (timeRange != null)
//						    stmtEl.setAttribute("TIMERANGE", timeRange.name());
//
//						ArrayList<String> entityIDs = new ArrayList<String>();
//
//						ArrayList<SS3Entity> ss3Entities = statementField.getSS3Entities();
//						for (SS3Entity ss3Entity: ss3Entities) {
//							entityIDs.add(ss3Entity.getID());
//						}
//
//						StringBuffer bf = new StringBuffer();
//
//						for (int counter = 0; counter < entityIDs.size(); counter++) {
//							bf.append(entityIDs.get(counter));
//							if (counter < entityIDs.size() - 1)
//								bf.append(", ");
//						}
//
//						stmtEl.setAttribute("ENTITY-IDS", bf.toString());
//						wEl.appendChild(stmtEl);
//					}
				}
			}
		}
		return root;
	}

	private Element createDatabaseElement(SS3Database db, Document document) {
		Element root = document.createElement("DATABASE");

		root.appendChild(createTracksElement(db, document));
		root.appendChild(createNotesElement(db, document));
		root.appendChild(createGlossWindowElement(db, document));
		// root.appendChild(createStatementsElement(db, document));
		return root;
	}

	private Element createTracksElement(SS3Database db, Document document) {
		Element root = document.createElement("TRACKS");
		ArrayList tracks = db.getTracksList();
		for (int i = 0; i < tracks.size(); i++) {
			SS3Track track = (SS3Track) tracks.get(i);

			Element tEl = document.createElement("TRACK");
			tEl.setAttribute("FID", "" + track.getId());
			ArrayList<SS3Entity> entities = track.getEntities();
			for (SS3Entity entity : entities) {
				Element eEl = document.createElement("ENTITY");
				eEl.setAttribute("ID", entity.getID());
				eEl.setAttribute("REF_ENTITY_ID", entity.getRefEntityId());
//				eEl.setAttribute("PARENT_ENTITY_ID", entity.getParentEntityID());	// see below - SEGMENT-TIER-UTTERANCE-REFS
				eEl.setAttribute("COMP_START_ENTITY_ID",
						entity.getStartCompoundEntityID());
				eEl.setAttribute("COMP_END_ENTITY_ID",
						entity.getEndCompoundEntityID());

				// eEl.setAttribute("VALUE_START_LOCKED", Util
				// .getBooleanAsString(entity.isValueStartTimeLocked()));
				// eEl.setAttribute("VALUE_END_LOCKED", Util
				// .getBooleanAsString(entity.isValueEndTimeLocked()));
				// eEl.setAttribute("INITIAL_HOLD_LOCKED", Util
				// .getBooleanAsString(entity
				// .isInitialHoldStartTimeLocked()));
				// eEl.setAttribute("FINAL_HOLD_LOCKED", Util
				// .getBooleanAsString(entity.isFinalHoldEndTimeLocked()));

				GlossEntityProperties properties = entity
						.getGlossEntityProperties();

				GestureEntityProperties gestureProperties = entity
						.getGestureEntityProperties();

				if (track.getId().equalsIgnoreCase(Util.DOM_HANDSHAPE_FIELD_ID)
						|| track.getId().equalsIgnoreCase(Util.NDOM_HANDSHAPE_FIELD_ID)
						|| track.getId().equalsIgnoreCase(Util.DOM_HAND_FIELD_ID)
						|| track.getId().equalsIgnoreCase(Util.NDOM_HAND_FIELD_ID)) {
					
					boolean locked = entity.isLockedGlossPhoneAlignments();
					if(locked) {
						eEl.setAttribute("LOCKED_GLOSS_PHON_ALIGNMENT", "T");
					} else {
						eEl.setAttribute("LOCKED_GLOSS_PHON_ALIGNMENT", "F");
					}
				} else if (track.getId().equalsIgnoreCase(Util.DOM_GESTURE_HANDSHAPE_FIELD_ID)
						|| track.getId().equalsIgnoreCase(Util.NDOM_GESTURE_HANDSHAPE_FIELD_ID)
						|| track.getId().equalsIgnoreCase(Util.DOM_GESTURE_HAND_FIELD_ID)
						|| track.getId().equalsIgnoreCase(Util.NDOM_GESTURE_HAND_FIELD_ID)) {
					
					boolean locked = entity.isLockedGesturePhoneAlignments();
					if(locked) {
						eEl.setAttribute("LOCKED_GESTURE_PHON_ALIGNMENT", "T");
					} else {
						eEl.setAttribute("LOCKED_GESTURE_PHON_ALIGNMENT", "F");
					}
				}

				if ((track.getId().equalsIgnoreCase(Util.DOM_GLOSS_FIELD_ID) || track.getId()
						.equalsIgnoreCase(Util.NDOM_GLOSS_FIELD_ID))) {
					if (properties != null) {

						Element pEl = document.createElement("PROPERTIES");
						pEl.setAttribute("SIGN", properties.getSignType() == null ? "" : properties.getSignType().name());
						pEl.setAttribute("NAME_SIGN", properties.getNameSign());
						pEl.setAttribute("PASSIVE_BASE_ARM", properties.getPassiveBaseArm());
						pEl.setAttribute("NO_OF_HANDS", properties.getNumberOfHands());

						if (track.getId().equals(Util.NDOM_GLOSS_FIELD_ID) && "2".equals(properties.getNumberOfHands())) {
                            entity.setDependentSignTypeEntityID(null);
                        }

						pEl.setAttribute("TWOHANDED_HANDSHAPES", properties.getTwoHandedHandshapesClassification());
						pEl.setAttribute("START_END_HANDSHAPES", properties.getStartEndHandshape());
						pEl.setAttribute("NDOM_HAND_START_END_HANDSHAPE", properties.getStartEndHandshapeNdomHand());
						pEl.setAttribute("DOM_HAND_START_END_HANDSHAPE",
								properties.getStartEndHandshapeDomHand());

						eEl.appendChild(pEl);

						eEl.setAttribute("MARKED_NUMBER_OF_HANDS",
								properties.getMarkedNumberOfHands());
						eEl.setAttribute("CORRESPONDING_GLOSS_ENTITY_ID",
								entity.getDependentGlossEntityID());
					}

					boolean lockedGlossPhoneAlignments = entity
							.isLockedGlossPhoneAlignments();
					boolean lockedLeftRightAlignments = entity
							.isLockedDomNdomAlignments();

					String lockedGlossPhonAlignment = lockedGlossPhoneAlignments ? "T"
							: "F";
					String lockedLeftRightAlignment = lockedLeftRightAlignments ? "T"
							: "F";

					String depGlossID = entity.getDependentGlossEntityID();
					String depHandID = entity.getDependentHandEntityID();
					String depHandShapeID = entity
							.getDependentHandShapeEntityID();
					String depLocationID = entity.getDependentLocationEntityID();
					String depSignTypeID = entity.getDependentSignTypeEntityID();
				
					eEl.setAttribute("LOCKED_DOM_NDOM_ALIGNMENT",
							lockedLeftRightAlignment);
					eEl.setAttribute("LOCKED_GLOSS_PHON_ALIGNMENT",
							lockedGlossPhonAlignment);

					eEl.setAttribute("DEPENDENT_GLOSS_ID", depGlossID);
					eEl.setAttribute("DEPENDENT_HAND_ID", depHandID);
					eEl.setAttribute("DEPENDENT_HAND_SHAPE_ID", depHandShapeID);
					eEl.setAttribute("DEPENDENT_LOCATION_ID", depLocationID);
					eEl.setAttribute("DEPENDENT_SIGN_TYPE_ID", depSignTypeID);
				} else if ( track.getId().equalsIgnoreCase(Util.DOM_GESTURE_FIELD_ID) || 
						track.getId().equalsIgnoreCase(Util.NDOM_GESTURE_FIELD_ID)) {
					
					boolean incompleteGesture = entity.isIncompleteGesture();
					if (incompleteGesture)
						eEl.setAttribute("INCOMPLETE", "T");
					else
						eEl.setAttribute("INCOMPLETE", "F");
					
					if (gestureProperties != null) {
						Element pEl = document.createElement("PROPERTIES");
						pEl.setAttribute("SIGN", gestureProperties.getSignType() == null ? "" : gestureProperties.getSignType().name());
						pEl.setAttribute("NAME_SIGN", gestureProperties.getNameSign());
						pEl.setAttribute("PASSIVE_BASE_ARM", gestureProperties.getPassiveBaseArm());
						pEl.setAttribute("NO_OF_HANDS", gestureProperties.getNumberOfHands());

						pEl.setAttribute("TWOHANDED_HANDSHAPES", gestureProperties.getTwoHandedHandshapesClassification());
						pEl.setAttribute("START_END_HANDSHAPES", gestureProperties.getStartEndHandshape());
						pEl.setAttribute("NDOM_HAND_START_END_HANDSHAPE", gestureProperties.getStartEndHandshapeNdomHand());
						pEl.setAttribute("DOM_HAND_START_END_HANDSHAPE",
								gestureProperties.getStartEndHandshapeDomHand());

						eEl.appendChild(pEl);

						eEl.setAttribute("DOMINANCE",
								gestureProperties.getDominamtHand());
						eEl.setAttribute("MARKED_NUMBER_OF_HANDS",
								gestureProperties.getMarkedNumberOfHands());
						eEl.setAttribute("CORRESPONDING_GLOSS_ENTITY_ID",
								entity.getDependentGlossEntityID());
					}

					boolean lockedGesturePhoneAlignments = entity
							.isLockedGesturePhoneAlignments();
					boolean lockedLeftRightAlignments = entity
							.isLockedDomNdomAlignments();

					String lockedGesturePhonAlignment = lockedGesturePhoneAlignments ? "T"
							: "F";
					String lockedLeftRightAlignment = lockedLeftRightAlignments ? "T"
							: "F";
					
					String depGestureID = entity.getDependentGestureEntityID();
					String depApexID = entity.getDependentApexEntityID();
					String depHandID = entity.getDependentHandEntityID();
					String depHandShapeID = entity.getDependentHandShapeEntityID();
					String depOrientationID = entity.getDependentOrientationEntityID();
					String depGestureTypeID = entity.getDependentSignTypeEntityID();
				
					eEl.setAttribute("LOCKED_DOM_NDOM_ALIGNMENT",
							lockedLeftRightAlignment);
					eEl.setAttribute("LOCKED_GESTURE_PHON_ALIGNMENT",
							lockedGesturePhonAlignment);
					
					eEl.setAttribute("DEPENDENT_GESTURE_ID", depGestureID);
					eEl.setAttribute("DEPENDENT_APEX_ID", depApexID);
					eEl.setAttribute("DEPENDENT_HAND_ID", depHandID);
					eEl.setAttribute("DEPENDENT_HAND_SHAPE_ID", depHandShapeID);
					eEl.setAttribute("DEPENDENT_ORIENTATION_ID", depOrientationID);
					eEl.setAttribute("DEPENDENT_GESTURE_TYPE_ID", depGestureTypeID);
					
				} else if (track.getId().equalsIgnoreCase(Util.DOM_GESTURE_TYPE_FIELD_ID)
						|| track.getId().equalsIgnoreCase(Util.NDOM_GESTURE_TYPE_FIELD_ID)) {
//					System.out.println(eEl);
				} else if (track.getId().equalsIgnoreCase(Util.FREE_TEXT_FIELD_ID)) {
					if (entity.isValidTime()) {
						eEl.setAttribute("VALID_TIME", "T");
					}
				}

				/**
				 * <ENTITY> <PROPERTIES
				 * SIGN="FINGERSPELLED/LOAN/LEXICAL/CLASSIFIER/GESTURE/NUMBER"
				 * NAME_SIGN="T/F" PASSIVE_BASE_ARM="T/F" NO_OF_HANDS="1/2"
				 *
				 * TWOHANDED_HANDSHAPES="SAME/ALTERNATING_SAME/DIFFERENT"
				 * START_END_HANDSHAPES="SAME/DIFFERENT"
				 * NDOM_HAND_START_END_HANDSHAPE="SAME/DIFFERENT"
				 * DOM_HAND_START_END_HANDSHAPE="SAME/DIFFERENT" />
				 * <DOMINANCE="L/R"/> <MARKED_NUMBER_OF_HANDS="T/F"/> </ENTITY>
				 */

				ArrayList<SS3Item> items = entity.getItemsAsList();
				int order = 0;
				for (SS3Item item : items) {
					Element iEl = document.createElement("ITEM");

					String vid = item.getId();
					iEl.setAttribute("VID", vid);
					iEl.setAttribute("IID", item.getImageID());

					if (item.isTextOnly()) {
						iEl.setAttribute("TEXT_ONLY", "T");
					}

					if (item.getLocationRegionID() != null && !item.getLocationRegionID().isEmpty()) {
						iEl.setAttribute("LOCATION_REGION_ID", item.getLocationRegionID());
						iEl.setAttribute("LOCATION_TYPE_ID", String.valueOf(item.getLocationTypeID()));
					}

					if (item.getNoteID() != null
							&& item.getNoteID().length() > 0) {
						iEl.setAttribute("NOTE_ID", item.getNoteID());
					}

                    addStartEndTimeAttributes(iEl, item.getStartTimeInfo(), item.getEndTimeInfo());
                    iEl.setAttribute("ORDER", String.valueOf(++order));

					if (item.getText() != null && item.getText().length() > 0) {

						// String str = item.getText();
						// System.out.println("TEXT = [" + str + "]");

						iEl.appendChild(document.createTextNode(item.getText()));
					}
					eEl.appendChild(iEl);
				}
				
				// parent utterances of this entity
				Element glossWindowItemRefsEl = document.createElement("SEGMENT-TIER-UTTERANCE-REFS");
				
				ArrayList<String> parentEntityIDs = entity.getParentEntityIDs();
				if (parentEntityIDs != null && ! parentEntityIDs.isEmpty()) {
					for (String parentEntityID: parentEntityIDs) {
						Element glossWindowItemRefEl = document.createElement("REF");
						write(glossWindowItemRefEl, "SEGMENT-TIER-UTTERANCE_ID", parentEntityID);
						glossWindowItemRefsEl.appendChild(glossWindowItemRefEl);
					}
				}
				
				eEl.appendChild(glossWindowItemRefsEl);
				
				tEl.appendChild(eEl);
			}

			ArrayList<SS3Item> nonCompleteEventsList = track
					.getNonCompleteEvents();
			if (nonCompleteEventsList.size() > 0) {
				Element neEl = document.createElement("NON_COMPLETE_EVENTS");
				for (int n = 0; n < nonCompleteEventsList.size(); n++) {
					SS3Item item = nonCompleteEventsList.get(n);

					Element iEl = document.createElement("ITEM");

                    addStartEndTimeAttributes(iEl, item.getStartTimeInfo(), item.getEndTimeInfo());

					String sf = item.isStartTimeFinal() ? "T" : "F";
					iEl.setAttribute("SF", sf);
					String ef = item.isEndTimeFinal() ? "T" : "F";
					iEl.setAttribute("EF", ef);
					if (item.getText() != null && item.getText().length() > 0) {
						iEl.appendChild(document.createTextNode(item.getText()));
					}

					iEl.setAttribute("COMP_START_ENTITY_ID",
							item.getStartCompoundEntityID());
					iEl.setAttribute("COMP_END_ENTITY_ID",
							item.getEndCompoundEntityID());

					neEl.appendChild(iEl);
				}
				tEl.appendChild(neEl);
			}

			ArrayList<SS3Item> enteredTxtList = track.getEnteredTextEvents();
			if (enteredTxtList.size() > 0) {
				Element etEl = document.createElement("ENTERED_TEXT");
				for (int t = 0; t < enteredTxtList.size(); t++) {
					SS3Item item1 = enteredTxtList.get(t);

					Element iEl = document.createElement("ITEM");

                    addStartEndTimeAttributes(iEl, item1.getStartTimeInfo(), item1.getEndTimeInfo());

					if (item1.getText() != null && item1.getText().length() > 0) {
						iEl.appendChild(document.createTextNode(item1.getText()));
					}
					etEl.appendChild(iEl);
				}
				tEl.appendChild(etEl);
			}
			root.appendChild(tEl);
		}
		return root;
	}

	private Element createNotesElement(SS3Database db, Document document) {
		Element root = document.createElement("NOTES");
		ArrayList<SS3Note> notes = db.getNotesList();
		for (SS3Note note : notes) {
			Element nEl = document.createElement("NOTE");
			nEl.setAttribute("ID", note.getID());
			nEl.setAttribute("CODER-ID", note.getCoderID());
			nEl.setAttribute("TIME_CREATED", note.getTimeCreated());
			nEl.setAttribute("TIME_UPDATED", note.getTimeUpdated());
			Element sEl = document.createElement("SUBJECT");
			sEl.appendChild(document.createTextNode(note.getSubject()));
			nEl.appendChild(sEl);
			Element bEl = document.createElement("BODY");
			bEl.appendChild(document.createTextNode(note.getBody()));
			nEl.appendChild(bEl);
			root.appendChild(nEl);
		}
		return root;
	}

	public void saveMediaLibrary(SS3MediaLibrary library) throws Exception {
		Document document = XMLWriter.createNewXMLDocument(library
				.getFileName());

		try {
			Element root = document.createElement("SIGNSTREAM-DATABASE");
			Element distributorElement = document.createElement("DISTRIBUTOR");
			distributorElement
					.appendChild(document
							.createTextNode("American Sign Language Linguistic Research Project, Boston University, Boston, MA."));
			// document.createTextNode(prf.getDistributor()));
			root.appendChild(distributorElement);

			Element searchLocationElement = document
					.createElement("SEARCH_LOCATIONS");
			root.appendChild(searchLocationElement);

			int priority = 1;
			ArrayList<SerchableDirectory> searchLocations = library
					.getDirectoriesToSearch();
			for (SerchableDirectory directory : searchLocations) {
				Element searchElement = document
						.createElement("SEARCH_LOCATION");
				File dirFile = new File(directory.getDirectory());
				File baseDir = new File(System.getProperty("user.dir"));
				String relativizedPath = Util.relativizePath(baseDir, dirFile);
				searchElement.setAttribute("PATH", relativizedPath);
				String searchable = directory.isSearchable() ? "Y" : "N";
				searchElement.setAttribute("LOOKUP", searchable);
				searchElement.setAttribute("PRIORITY", "" + priority);
				int numberOfSearches = directory.isSearchable() ? directory
						.getNumberOfSearches() + 1 : directory
						.getNumberOfSearches();
				searchElement.setAttribute("SEARCH_COUNT", ""
						+ numberOfSearches);
				priority++;
				searchLocationElement.appendChild(searchElement);
			}

			XMLWriter.saveDocument(root, library.getFileName());

		} catch (Exception ex) {

			SS3SignStreamApplication.logger.log(Level.INFO,
					"Media library [" + library.getFileName()
							+ "] had NOT been saved. " + ex.getStackTrace());

			// ex.printStackTrace();

			throw new Exception("Media library [" + library.getFileName()
					+ "] had NOT been saved.");
		}
	}

	public boolean deleteDatabase(SS3Database db) {
		String directory = "SignStreamDatabases" + File.separator
				+ db.getDirectory();
		String fileName = db.getFileName();

		if (db.getReadFromXMLFile() != null) {
			fileName = db.getReadFromXMLFile().getPath();
		} else {
			fileName = directory + File.separator + db.getFileName();
		}

		File file = new File(fileName);
		return file.delete();
	}

	/**
	 * Saves the database to a temporary directory.
	 * Sets the database's file extension if it hasn't been set.
	 * If saving a database that has the extension "[...].ss3.ss3" you must set the file extension
	 * yourself before passing it to this function.
	 * @param db
	 * @return True if the files were saved to the temporary directory successfully and False
	 * otherwise.
	 * @throws Exception
	 */
	public boolean writeDatabase(SS3Database db) throws Exception {
		String dbDirectory = db.getParentDirectoryPath();

		String dir = db.getDirectory();

		if(!dir.startsWith(SS3Database.tempDirectoryPrefix))
    		dir = SS3Database.tempDirectoryPrefix + dir;
    	if(!dir.endsWith(SS3Database.COLLECTION_EXT))
    		dir += SS3Database.COLLECTION_EXT;
    	db.setDirectory(dir);
		
		String directory = dbDirectory + File.separator + dir;


		SS3SignStreamApplication.logger.log(Level.INFO, "Saving database ["
				+ directory + "]...");

		String collectionName = SS3Singleton.getSignStreamApplication().getTitle();

		File databasesFile = new File(dbDirectory);
		if (!databasesFile.exists()) {
			try {
				boolean createDirectory = databasesFile.mkdir();

				if (!createDirectory) {
					String[] str = new String[4];
					str[0] = "Collection " + collectionName
							+ " could not be saved ";
					str[1] = "because its parent directory could not be created.";
					str[2] = "Please make sure that only legal characters are used.";
					str[3] = "Also try saving the collection in a different directory.";

					SS3Singleton.getErrorMessages()
							.showSaveCollectionErrorMessage(str);
					return false;
				}
			} catch (SecurityException sEx) {
				String[] str = new String[4];
				str[0] = "Collection " + collectionName
						+ " could not be saved ";
				str[1] = "The Security manager does not permit the named";
				str[2] = "directory to be created.  Please verify write";
				str[3] = "permissions and try again.";

				SS3Singleton.getErrorMessages().showSaveCollectionErrorMessage(
						str);
				return false;
			}
		}

		File directoryFile = new File(directory);

		if (!directoryFile.exists()) {
			try {
				boolean createFile = directoryFile.mkdir();
				if (!createFile) {

					String[] str = new String[4];
					str[0] = "Collection " + collectionName
							+ " could not be saved ";
					str[1] = "because collection directory was not created. ";
					str[2] = "Please make sure that only legal characters had been used.";
					str[3] = "Also try saving the collection in different directory.";

					SS3Singleton.getErrorMessages()
							.showSaveCollectionErrorMessage(str);
					return false;
				}
			} catch (SecurityException sEx) {

				String[] str = new String[4];
				str[0] = "Collection " + collectionName
						+ " could not be saved. ";
				str[1] = "The Security manager does not permit the named";
				str[2] = "directory to be created.  Please verify write";
				str[3] = "permissions and try again.";

				SS3Singleton.getErrorMessages().showSaveCollectionErrorMessage(
						str);
				return false;
			}
		}

		String name0 = db.getFileName();

		int t = name0.indexOf(SS3Database.macroFileExtension);
		if (t == -1) {
			t = name0.indexOf(".xml");
		}

		String name = name0;
		if (t > -1) {
			name = name0.substring(0, t);
		}

		// check if filename is valid
		String newFileName = "";
		char[] chars = name.toCharArray();

		for (int i = 0; i < chars.length; i++) {
			char c = chars[i];
			if (Character.isLetterOrDigit(c)) {
				newFileName = newFileName + c;
			} else if (Character.isWhitespace(c)) {
				newFileName.trim();
				newFileName = newFileName + c;
			} else {
				String str = new String("" + c);
				if (str.equalsIgnoreCase("-") || str.equalsIgnoreCase("+")) {
					newFileName = newFileName + str;
				} else if (str.equalsIgnoreCase("#")) {
					newFileName = newFileName + "loan";
				} else if (str.equalsIgnoreCase(File.pathSeparator)) {
					newFileName = newFileName + "or";
				}
			}
		}
		
		String macroUnit = directory + File.separator + SS3Database.COLLECTION_MACRO_UNIT_DIR;
		File macroUnitDir = new File(macroUnit);

		if (!macroUnitDir.exists()) {
			try {
				boolean createFile = macroUnitDir.mkdir();
				if (!createFile) {

					String[] str = new String[4];
					str[0] = "Collection " + collectionName
							+ " could not be saved ";
					str[1] = "because collection directory was not created. ";
					str[2] = "Please make sure that only legal characters had been used.";
					str[3] = "Also try saving the collection in different directory.";

					SS3Singleton.getErrorMessages()
							.showSaveCollectionErrorMessage(str);
					return false;
				}
			} catch (SecurityException sEx) {

				String[] str = new String[4];
				str[0] = "Collection " + collectionName
						+ " could not be saved. ";
				str[1] = "The Security manager does not permit the named";
				str[2] = "directory to be created.  Please verify write";
				str[3] = "permissions and try again.";

				SS3Singleton.getErrorMessages().showSaveCollectionErrorMessage(
						str);
				return false;
			}
		}

		String fileName = macroUnit + File.separator + newFileName + SS3Database.macroFileExtension;

		File file = new File(fileName);

		if(file.exists()) {
			file.delete();
		}

		Document document = null;
		try {
			document = XMLWriter.createNewXMLDocument(fileName);
		} catch (Exception ex) {

			String[] str = new String[4];
			str[0] = "Collection " + collectionName + " could not be saved. ";
			str[1] = ex.getMessage();
			str[2] = "";
			str[3] = "";

			SS3Singleton.getErrorMessages().showSaveCollectionErrorMessage(str);
			return false;
		}

		try {
			SS3DBProfile prf = db.getProfile();
			Element root = document.createElement("SIGNSTREAM-DATABASE");
			root.setAttribute("SIGNSTREAM-VERSION",
					"" + prf.getSignStreamVersion());
//			root.setAttribute("DATABASE-VERSION", db.isBroken() ? "3.0" : "3.1");
			root.setAttribute("DATABASE-VERSION", "3.1");
			root.setAttribute("EXCERPT", db.getExcerpt());
			root.setAttribute("EXCERPT_NUMBER", db.getNumber());

            root.appendChild(createVideosElement(db, document));
            root.appendChild(createGraphsElement(db, document));

			root.appendChild(createUtilsElement(db, document));
			root.appendChild(createDistributorElement(prf, document));
			root.appendChild(createReferencesElement(db, document));
			// root.appendChild(createAuthorsElement(db, document));
			root.appendChild(createMediaFilesOnScreenLocationElement(db,
					document));
			root.appendChild(createCodingSchemeElement(db, document));
			root.appendChild(createDatabaseElement(db, document));

			boolean savedCoolection = XMLWriter.saveDocument(root, fileName);
			if (!savedCoolection) {

				String[] str = new String[4];
				str[0] = "Collection " + collectionName
						+ " could not be saved. ";
				str[1] = "Please manually delete the directory and/or file";
				str[2] = "with the collection name and try again.";
				str[3] = "";

				SS3Singleton.getErrorMessages().showSaveCollectionErrorMessage(
						str);
				return false;
			}

			// showCollectionSavedMessage();
			return true;
		} catch (Exception ex) {
			SS3SignStreamApplication.logger.log(Level.INFO, "Collection ["
					+ dbDirectory
					+ "] had NOT been saved. Error in Macro Unit [" + fileName
					+ "]" + ex.getStackTrace());

			String[] str = new String[4];
			str[0] = "Collection " + collectionName + " could not be saved. ";
			str[1] = ex.getMessage();
			str[2] = "";
			str[3] = "";

			SS3Singleton.getErrorMessages().showSaveCollectionErrorMessage(str);
			ex.printStackTrace();
			return false;
		}
	}

	private JPopupMenu collectionSavedPopup = null;
	private JLabel label = new JLabel("Collection had been saved.");

	private void showCollectionSavedMessage() {

		if (collectionSavedPopup != null) {
			UserInterfaceUtil.setLabelLNF(label);
			collectionSavedPopup = new JPopupMenu();

			final JButton okBtn = new JButton("OK");

			JPanel msgPanel = new JPanel();
			GridBagConstraints gbc = new GridBagConstraints();

			msgPanel.setLayout(new GridBagLayout());
			gbc.anchor = GridBagConstraints.WEST;

			gbc.gridwidth = 3;
			gbc.gridx = 0;
			gbc.gridy = 0;
			msgPanel.add(label, gbc);

			gbc.gridy = 1;
			gbc.gridx = 0;
			msgPanel.add(new JLabel(" "), gbc);

			gbc.gridwidth = 1;
			gbc.anchor = GridBagConstraints.EAST;
			gbc.gridy = 3;
			gbc.gridx = 0;
			msgPanel.add(new JLabel(" "), gbc);
			gbc.gridx = 2;
			msgPanel.add(okBtn, gbc);

			okBtn.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					collectionSavedPopup.setVisible(false);
				}
			});

			SwingUtilities.invokeLater(new Runnable() {
				@Override
				public void run() {
					okBtn.requestFocus();
				}
			});

			Dimension d = new Dimension();
			int width = msgPanel.getPreferredSize().width + 30;
			int height = msgPanel.getPreferredSize().height + 20;

			d.setSize(width, height);

			collectionSavedPopup.add(msgPanel);
			collectionSavedPopup.setPopupSize(d);
		}

		JPanel panel = SS3Singleton.getSignStreamApplication()
				.getSignStreamSegmentPanel().getUtteranceView();
		// popup.show(field.getSignStreamSegmentPanel().getUtteranceView(), 250,
		// 250);
		if (panel != null) {
			panel = SS3Singleton.getSignStreamApplication()
					.getSignStreamSegmentPanel().getDatabaseView();
		}
		collectionSavedPopup.show(panel, 250, 250);
		collectionSavedPopup.setVisible(true);
	}
}
