package edu.bu.signstream.common.db.ss3;

import static edu.bu.signstream.common.db.ss4.SS4Util.readChildOpt;
import static edu.bu.signstream.common.db.ss4.SS4Util.readChildren;
import static edu.bu.signstream.common.db.ss4.SS4Util.readLongOpt;
import static edu.bu.signstream.common.db.ss4.SS4Util.readOpt;
import static edu.bu.signstream.common.db.ss4.SS4Util.write;
import static edu.bu.signstream.common.db.ss4.SS4Util.writeLong;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map.Entry;
import java.util.stream.Collectors;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import edu.bu.signstream.common.files.SSResources;
import edu.bu.signstream.common.util.IntervalSet;
import edu.bu.signstream.common.util.Util;
import edu.bu.signstream.common.util.vo.Constants;
import edu.bu.signstream.common.util.vo.ss3.codingScheme.SS3CodingScheme;
import edu.bu.signstream.common.util.vo.ss3.codingScheme.SS3Field;
import edu.bu.signstream.common.util.vo.ss3.codingScheme.SS3Field.Alignment;
import edu.bu.signstream.common.util.vo.ss3.codingScheme.SS3FieldValue;
import edu.bu.signstream.grepresentation.fields.nonManualField.GlossEntityProperties;
import edu.bu.signstream.grepresentation.reguestsHandler.SS3Singleton;
import edu.bu.signstream.time.Bound;
import edu.bu.signstream.time.Rounding;
import edu.bu.signstream.time.TimeRange;
import edu.bu.signstream.time.TimeUnit;

import static edu.bu.signstream.common.db.ss4.SS4Util.*;

public class SS3_1Converter extends Converter {
	
	SS3Validator validator = null;
	
	Document doc = null;
	
	Element rootEl = null, dbEl = null, glossWindowEl = null;
	
	Long timestep, duration;
	
	TimeUnit frameTimeUnit = null;
	
	// Entity ID - Track ID
	HashMap<String, String> entityTrackMap = new HashMap<String, String>();

	HashMap<String, String> deletedEntities = new HashMap<String, String>();
	
	HashMap<String, Element> partitionElements = new HashMap<String, Element>();
	HashMap<String, Element> tierElements = new HashMap<String, Element>();
	HashMap<String, Element> utteranceElements = new HashMap<String, Element>();

	HashMap<String, String> tierPartitionMap = new HashMap<String, String>();		// map: segmentTierID -> temporalPartitionID
	HashMap<String, String> utteranceTierMap = new HashMap<String, String>();	// map: utteranceID -> segmentTierID
	HashMap<String, String> entityUtteranceMap = new HashMap<String, String>();	// map: entityID -> utteranceID
	
	HashMap<String, Element> trackElements = new HashMap<String, Element>();
	HashMap<String, Element> entityElements = new HashMap<String, Element>();
	// for itemElements: key is entity id, value is list of items in entity keyed by its item text
	HashMap<String, ArrayList<Element>> itemElements = new HashMap<String, ArrayList<Element>>();

	HashMap<String, IntervalSet<String>> utteranceIntervals = new HashMap<String, IntervalSet<String>>();	// key is segmentTierID
	
//	HashMap<String, HashMap<String, Element>> localSchemeFieldElements = new HashMap<String, HashMap<String, Element>>();
	
	@Override
	public SS3ValidatorConverterReturns convert(Validator validator) {
		if (validator instanceof SS3Validator)
			return convert((SS3Validator) validator);
		
		return null;
	}

	public SS3ValidatorConverterReturns convert(SS3Validator validator) {
		this.validator = validator;
		
		returns = new SS3ValidatorConverterReturns();
        returns.errors.clear();
        returns.modifications.clear();
		
    	this.doc = validator.doc;
    	
    	this.rootEl = validator.rootEl;
    	this.dbEl = validator.dbEl;
    	this.glossWindowEl = validator.glossWindowEl;
    	
    	this.timestep = validator.timestep;
    	this.duration = validator.duration;
    	this.frameTimeUnit = validator.frameTimeUnit;
    	
    	this.entityTrackMap = validator.entityTrackMap;
    	this.deletedEntities = validator.deletedEntities;
    	
    	this.partitionElements = validator.partitionElements;
    	this.tierElements = validator.tierElements;
    	this.utteranceElements = validator.utteranceElements;
    	
    	this.tierPartitionMap = validator.tierPartitionMap;
    	this.utteranceTierMap = validator.utteranceTierMap;
    	this.entityUtteranceMap = validator.entityUtteranceMap;
    	
    	this.trackElements = validator.trackElements;
    	this.entityElements = validator.entityElements;
    	this.itemElements = validator.itemElements;
    	
    	this.utteranceIntervals = validator.utteranceIntervals;
    	
//    	this.localSchemeFieldElements = validator.localSchemeFields;
		convertVersions();
		
		convertAddMediaOrderAttribute();
		
		convertAddItemOrderAttribute();
		
		convertGlossAttributeMisspelings();
		
		validateAndCorrectLocalCodingScheme();
		
		convertGlossWindowToWindow();
		
		convertTemporalPartitionSegmentsToSegmentTiers();
		
		convertGlossWindowItemsToSegmentTierUtterances();
		
		convertParentEntityRefsToElement();
		
		removeBlankLocationEntities();
		
		convertTimestampsToFrames();
		
		convertPhaseCodes();
		
		convertIsDefaultSegmentTierFormat();
		
		return returns;
	}
	
	private void convertVersions() {
		rootEl.setAttribute("SIGNSTREAM-VERSION", "3.4.0");
		addToModifications(0, "SIGNSTREAM-VERSION was updated to 3.4.0", null);
		
		rootEl.setAttribute("DATABASE-VERSION", "3.1");
		addToModifications(0, "DATABASE-VERSION was updated to 3.1", null);
		// removed this, because it required checking whether we were formerly in 3.0 
		// in SS3SignStreamApplication.saveDatabaseToFile()
		// GD 2023-04-19 - putting the version update to 3.1 back in, because in DatabaseDelegate we now want to:
		// validation in 3.0, convert 3.0 to 3.1, validate in 3.1
		// and 3.1 validation requires the version to be 3.1
	}
	
	private void convertAddMediaOrderAttribute() {
        var mediaFilesEl = readChildOpt(rootEl, "MEDIA-FILES").orElse(null);

        ArrayList<Element> list = readChildren(mediaFilesEl, "MEDIA-FILE").collect(Collectors.toCollection(ArrayList::new));
        for (int i = 0; i < list.size(); i++) {
    		var order = "" + (i + 1);
        		Element mEl = list.get(i);
        		if ( ! mEl.hasAttribute("ORDER") || ! mEl.getAttribute("ORDER").equals(order)) {
            		mEl.setAttribute("ORDER", order);
            		
            		addToModifications(0, "Attribute ORDER was added to Media File {ID=" + readOpt(mEl, "ID").get() +  " }", null);
        		}
        }
	}

	private void convertAddItemOrderAttribute() {
		var domTrack = trackElements.get(Util.DOM_GLOSS_FIELD_ID);
    	var ndomTrack = trackElements.get(Util.NDOM_GLOSS_FIELD_ID);
    	if (domTrack == null && ndomTrack == null)
    		return;
    	
    	var glossTrackElements = List.of(domTrack, ndomTrack);
    	
    	boolean madeModification = false;
        for (var trackEl : glossTrackElements) {
        	var trackID = readOpt(trackEl, "FID").orElse(null);
            var hsTrackId = trackID.equals(Util.DOM_GLOSS_FIELD_ID)
                ? Util.DOM_HANDSHAPE_FIELD_ID
                : Util.NDOM_HANDSHAPE_FIELD_ID;

            var locTrackID = trackID.equals(Util.DOM_GLOSS_FIELD_ID)
            		? Util.DOM_LOCATION_FIELD_ID
            		: Util.NDOM_LOCATION_FIELD_ID;

            var hsTrackEl = trackElements.get(hsTrackId);
            var locTrackEl = trackElements.get(locTrackID);

            for (Element entityEl: readChildren(trackEl, "ENTITY")
            		.collect(Collectors.toCollection(ArrayList::new))) {
                var eid = readOpt(entityEl, "ID").orElse(null);

            	var dependentHandshapeID = readOpt(entityEl, "DEPENDENT_HAND_SHAPE_ID").orElse(null);
            	var dependentLocationID = readOpt(entityEl, "DEPENDENT_LOCATION_ID").orElse(null);
            	
            	Element hsEl = null, locEl = null;
            	
            	for (Element hsEntityEl: readChildren(hsTrackEl, "ENTITY").collect(Collectors.toCollection(ArrayList::new))) {
            		var hsEID = readOpt(hsEntityEl, "ID").orElse(null);
            		if (hsEID != null && hsEID.equals(dependentHandshapeID)) {
            			hsEl = hsEntityEl;
            			break;
            		}
            	}
            	if (hsEl != null) {
                	var hsID = readOpt(hsEl, "ID").orElse(null);
                	
                	var hsItemsList = itemElements.get(hsID).stream().sorted(validator.comp).collect(
                			Collectors.toCollection(ArrayList<Element>::new));
                	
                	for (int i = 0; i < hsItemsList.size(); i++) {
                		var order = "" + (i + 1);
                		Element hsItem = hsItemsList.get(i);
                		if ( ! hsItem.hasAttribute("ORDER") || ! hsItem.getAttribute("ORDER").equals(order)) {
                    		write(hsItem, "ORDER", order);
                    		madeModification = true;
                		}
                	}
            	}
            	
            	for (Element locEntityEl: readChildren(locTrackEl, "ENTITY").collect(Collectors.toCollection(ArrayList::new))) {
            		var locEID = readOpt(locEntityEl, "ID").orElse(null);
            		if (locEID != null && locEID.equals(dependentLocationID)) {
            			locEl = locEntityEl;
            			break;
            		}
            	}
            	if (locEl != null) {
                	var locID = readOpt(locEl, "ID").orElse(null);
                	
                	var locItemsList = itemElements.get(locID).stream().sorted(validator.comp).collect(
                			Collectors.toCollection(ArrayList<Element>::new));
                	
                	for (int i = 0; i < locItemsList.size(); i++) {
                		var order = "" + (i + 1);
                		Element locItem = locItemsList.get(i);
                		if ( ! locItem.hasAttribute("ORDER") || ! locItem.getAttribute("ORDER").equals(order)) {
                    		write(locItem, "ORDER", order);
                			madeModification = true;
                		}
                	}
            	}
            }
        }
        
        if (madeModification)
        		addToModifications(0, "Attribute ORDER was added to all Handshape and Location entity items.", null);
	}
	
	private void convertGlossAttributeMisspelings() {
		var domTrack = trackElements.get(Util.DOM_GLOSS_FIELD_ID);
    	var ndomTrack = trackElements.get(Util.NDOM_GLOSS_FIELD_ID);
    	if (domTrack == null && ndomTrack == null)
    		return;
    	
    	var glossTrackElements = List.of(domTrack, ndomTrack);
    	
        	boolean madeLeftRightDomNdomModification = false;
        	boolean madeTwoHandedModification = false;
        	
        for (var trackEl : glossTrackElements) {
            for (Element entityEl: readChildren(trackEl, "ENTITY")
            		.collect(Collectors.toCollection(ArrayList::new))) {
                
                var elLocked = readOpt(entityEl, "LOCKED_LEF_RIGHT_ALIGNMENT").orElse(null);
                if (elLocked != null && !elLocked.isBlank()) {
                	entityEl.removeAttribute("LOCKED_LEF_RIGHT_ALIGNMENT");
                	write(entityEl, "LOCKED_DOM_NDOM_ALIGNMENT", elLocked);
                	madeLeftRightDomNdomModification = true;
                }

                var propertiesEl = readChildOpt(entityEl, "PROPERTIES").orElse(null);
                var twoHandedHandshapes = readOpt(propertiesEl, "TWOHANDED_HADSHAPES").orElse(null);
                if (twoHandedHandshapes != null && !twoHandedHandshapes.isBlank()) {
                	propertiesEl.removeAttribute("TWOHANDED_HADSHAPES");
                	
                    String newTwoHandedHandshapes = null;
            		if (twoHandedHandshapes.equals("SAME HANDSHAPES LR")) {
            			newTwoHandedHandshapes = GlossEntityProperties.SAME_HANDSHAPES_DOM_NDOM;
            		} else if (twoHandedHandshapes.equals("ALTERNATING SAME HANDSHAPES LR")) {
            			newTwoHandedHandshapes = GlossEntityProperties.ALTERNATING_SAME_HANDSHAPES_DOM_NDOM;
            		} else if (twoHandedHandshapes.equals("DIFFERENT HANDSHAPES LR")) {
            			newTwoHandedHandshapes = GlossEntityProperties.DIFFERENT_HANDSHAPES_DOM_NDOM;
            		}
            		
                	write(propertiesEl, "TWOHANDED_HANDSHAPES", newTwoHandedHandshapes);
                	madeTwoHandedModification = true;
                }
                
                var leftHandStartEndHandshape = readOpt(propertiesEl, "LEFT_HAND_START_END_HANDSHAPE").orElse("");
                write(propertiesEl, "NDOM_HAND_START_END_HANDSHAPE", leftHandStartEndHandshape);
                
                var rightHandStartEndHandshape = readOpt(propertiesEl, "RIGHT_HAND_START_END_HANDSHAPE").orElse("");
                write(propertiesEl, "DOM_HAND_START_END_HANDSHAPE", rightHandStartEndHandshape);
            }
        }

        if (madeLeftRightDomNdomModification)
        	addToModifications(0, "Attribute LOCKED_LEF_RIGHT_ALIGNMENT was corrected to LOCKED_DOM_NDOM_ALIGNMENT globally.", null);

        if (madeTwoHandedModification)
        	addToModifications(0, "Attribute TWOHANDED_HADSHAPES was corrected to TWOHANDED_HANDSHAPES globally.", null);

        addToModifications(0, "Attribute RIGHT_HAND_START_END_HANDSHAPE was corrected to DOM_HAND_START_END_HANDSHAPE globally.", null);

        addToModifications(0, "Attribute LEFT_HAND_START_END_HANDSHAPE was corrected to NDOM_HAND_START_END_HANDSHAPE globally.", null);
	}
    
    private void validateAndCorrectLocalCodingScheme() {
        var localSchemeEl = readChildOpt(rootEl, "LOCAL-CODING-SCHEME").orElse(null);
//        localSchemeFieldElements = new HashMap<String, HashMap<String, Element>>();

		String dv = System.getProperty("file.separator", "/");
		var path = SSResources.xmlPath + dv;
		var scPath = path + "3.3.5.defCodingScheme.xml";
        var reader = new SS3Reader();
        var defScheme3_3_5 = reader.parseSS3_3_5CodingScheme(scPath);	// using version from 3.3.5, as there were changes since then
        var defScheme = SS3Singleton.getSS3CodingSchemeCollection().getDefaultCodingScheme();
        
        var localScheme = new SS3CodingScheme();

        int maxOrder = defScheme.getMaxOrderAcrossAllFields();	// need to use the newer scheme for this, since ORDER attributes were not on fields in 3.3.5

        if (expect(localSchemeEl != null, 10034, "Missing LOCAL-CODING-SCHEME element.", null)) {
    		for (Element fEl: readChildren(localSchemeEl, "FIELD").collect(Collectors.toCollection(ArrayList::new))) {
                var repr = "Local coding scheme field " + repr(fEl, "ID");
                var id = readOpt(fEl, "ID").orElse(null);

    			var type = readOpt(fEl, "TYPE").orElse(null);
    			if ( ! Constants.EDITABLE_TYPE_VALUES.contains(type)) {
//    				localSchemeEl.removeChild(fEl);		// entire scheme element will be rebuilt below, so by continuing here we are not going to write it out in that new scheme element
    				addToModifications(0, repr + " was removed from the file since it is not an editable field.", null);
    				continue;
    			}

            	SS3Field defaultField3_3_5 = defScheme3_3_5.getField(id);
    			SS3Field localField = null;
    			
    			if (defaultField3_3_5 != null) {
    				localField = defaultField3_3_5.cloneField();
    				
    				var defaultField = defScheme.getField(id);
    				if (defaultField != null) {
    					localField.setOrder(defaultField.getOrder());
    				}
    			} else {
    				localField = new SS3Field();
    				localField.setId(id);
    				localField.setOrder(++maxOrder);
    			}

    			var fieldName = readOpt(fEl, "NAME").orElse(null);
    			var fieldLabel = readOpt(fEl, "LABEL").orElse(null);
    			var fieldPrefix = readOpt(fEl, "PREFIX").orElse(null);
    			var fieldAlignment = readOpt(fEl, "ALIGNMENT").orElse(null);
    			var fieldGroup = readOpt(fEl, "GROUP").orElse(null);
//    			var color = readOpt(fEl, "COLOR").orElse(null);		// ignore color, it seems always incorrect in old files

    			localField.setId(id);
    			localField.setName(fieldName);
    			localField.setLabel(fieldLabel);
    			localField.setPrefix(fieldPrefix);
    			localField.setType(type);
//    			localField.setColor(color);

    			if (fieldAlignment != null) {
    				if (fieldAlignment.equals("RIGHT"))
    					localField.setAlignment(Alignment.RIGHT);
    				else if (fieldAlignment.equals("LEFT"))
    					localField.setAlignment(Alignment.LEFT);
    				else if (fieldAlignment.equals("RIGHT_LEFT"))
    					localField.setAlignment(Alignment.RIGHT);
    				else if (fieldAlignment.equals("CENTER"))
    					localField.setAlignment(Alignment.CENTER);
    			} else
    				localField.setAlignment(null);

    			if (fieldGroup != null) {
    				String origGroup = null;
    				if (localField.getGroup() != null && ! localField.getGroup().equals(fieldGroup)) {
    					origGroup = localField.getGroup();
    				}
    				
    				switch (fieldGroup) {
    				case Constants.HEAD_GROUP_NAME_CAPS:
    					localField.setGroupID(Constants.HEAD_GROUP_ID);
    					localField.setGroup(Constants.HEAD_GROUP_NAME_CAPS);
    					break;
    				case Constants.FACE_GROUP_NAME_CAPS:
    					localField.setGroupID(Constants.FACE_GROUP_ID);
    					localField.setGroup(Constants.FACE_GROUP_NAME_CAPS);
    					break;
    				case Constants.BODY_GROUP_NAME_CAPS:
    					localField.setGroupID(Constants.BODY_GROUP_ID);
    					localField.setGroup(Constants.BODY_GROUP_NAME_CAPS);
    					break;
    				case Constants.GRAMMATICAL_GROUP_NAME_CAPS:
    					localField.setGroupID(Constants.GRAMMATICAL_GROUP_ID);
    					localField.setGroup(Constants.GRAMMATICAL_GROUP_NAME_CAPS);
    					break;
    				}
    				
    				if (origGroup != null && ! localField.getGroup().equals(origGroup)) {
    					addToModifications(0, repr + " had Group value updated from " + origGroup + " to " + localField.getGroup(), null);
    				}
    			}
    			
    			ArrayList<SS3FieldValue> fieldValues = new ArrayList<SS3FieldValue>();
    			
                var valueElements = new HashMap<String, Element>();
                readChildren(fEl, "VALUE").forEach(vEl -> {
                    var vrepr = repr + " value " + repr(vEl, "ID");
                    var vid = readOpt(vEl, "ID").orElse(null);
                    
                    // TODO: Validate IID is in the coding scheme
                    readOpt(vEl, "IID").ifPresent(iid -> {});

                    valueElements.put(vid, vEl);
                    
                    SS3FieldValue fieldValue = new SS3FieldValue();

                    var valueName = readOpt(vEl, "NAME").orElse(null);
                    var valueLabel = readOpt(vEl, "LABEL").orElse(null);
                    var valueIID = readOpt(vEl, "IID").orElse(null);
                    var valueType = readOpt(vEl, "TYPE").orElse(null);

                    fieldValue.setId(vid);
                    fieldValue.setName(valueName);
                    fieldValue.setLabel(valueLabel);
                    fieldValue.setIid(valueIID);
                    fieldValue.setType(valueType);
                    fieldValue.setParentFieldID(id);

                    fieldValues.add(fieldValue);
                });
                
                localField.setValues(fieldValues);
                
                if (defaultField3_3_5 != null && localField != null && defaultField3_3_5.equals(localField)) {
//    				localSchemeEl.removeChild(fEl);		// entire scheme element will be rebuilt below
    				addToModifications(0, repr + " was removed from the file since it has no differences from the default coding scheme field.", null);
    				continue;
                }
    			
    			var localGroup = localField.getGroup();
    			if (localGroup == null || localGroup.isBlank()) {
    				addToErrors(10036, repr + " missing GROUP.", null);
    			}

//                localSchemeFieldElements.put(id, values);
    			localScheme.addField(localField);

    			for (SS3FieldValue val: fieldValues) {
    				var vEl = valueElements.get(val.getId());
    				var vrepr = repr + " value " + repr(vEl, "ID");
    				addToModifications(0, vrepr + " was converted to the 3.1 database compatible version.", null);
    			}
    			addToModifications(0, repr + " was converted to the 3.1 database compatible version.", null);
    		}
        }
        
        // rebuild local coding scheme
        Element newLocalSchemeEl = SS3Writer.createCodingSchemeElement(localScheme, doc);
        rootEl.insertBefore(newLocalSchemeEl, localSchemeEl);
        rootEl.removeChild(localSchemeEl);
        addToModifications(0, "Local coding scheme completely updated to the 3.1 database version.", null);
    }
    
    private void convertGlossWindowToWindow() {
    	doc.renameNode(glossWindowEl, null, "WINDOW");
    }
    
    private void convertTemporalPartitionSegmentsToSegmentTiers() {
    	var partitionsEl = readChildOpt(glossWindowEl, "TEMPORAL-PARTITIONS").orElse(null);
    	readChildren(partitionsEl, "TEMPORAL-PARTITION").forEach(partitionEl -> {
    		for (Element tierEl: readChildren(partitionEl, "TEMPORAL-PARTITION-SEGMENT").collect(Collectors.toCollection(ArrayList::new))) {
				doc.renameNode(tierEl, null, "SEGMENT-TIER");
			}
    	});
    }
    
    private void convertGlossWindowItemsToSegmentTierUtterances() {
    	var partitionsEl = readChildOpt(glossWindowEl, "TEMPORAL-PARTITIONS").orElse(null);
    	readChildren(partitionsEl, "TEMPORAL-PARTITION").forEach(partitionEl -> {
    		readChildren(partitionEl, "SEGMENT-TIER").forEach(tierEl -> {
    			
    			ArrayList<String> entityIDs = new ArrayList<String>();
    			HashMap<String, ArrayList<String>> list = new HashMap<String, ArrayList<String>>();	// fieldID -> list of entity-ids
    			
    			for (Element utteranceEl: readChildren(tierEl, "GLOSS-WINDOW-ITEM").collect(Collectors.toCollection(ArrayList::new))) {
    				doc.renameNode(utteranceEl, null, "SEGMENT-TIER-UTTERANCE");
    				
    				for (Element stmtFieldEl: readChildren(utteranceEl, "STATEMENT-FIELD").collect(Collectors.toCollection(ArrayList::new))) {
    					var fieldID = stmtFieldEl.getAttribute("FIELD-ID");
        				var utteranceEntityIDs = stmtFieldEl.getAttribute("ENTITY-IDS");
        				
        				if (isNotBlank(utteranceEntityIDs)) {
        					if (list.get(fieldID) == null) {
        						list.put(fieldID, new ArrayList<String>());
        					}
        					
        					ArrayList<String> fieldEntityIDs = list.get(fieldID);
        					
            				String[] eidList = utteranceEntityIDs.split(", ");
            				for (String eid: eidList) {
            					fieldEntityIDs.add(eid);
            				}
        				}
        				
        				utteranceEl.removeChild(stmtFieldEl);
    				}
    				
    			}
    			
    			for (String fieldID: list.keySet()) {
    				ArrayList<String> fieldEntityIDs = list.get(fieldID);

        			var statementFieldEl = doc.createElement("STATEMENT-FIELD");
        			statementFieldEl.setAttribute("ENTITY-IDS", String.join(", ", fieldEntityIDs));
        			statementFieldEl.setAttribute("FIELD-ID", fieldID);
        			tierEl.appendChild(statementFieldEl);
    			}

    		});
    		
    	});
    }
    
    private void convertParentEntityRefsToElement() {
    	for(Entry<String, Element> entry: entityElements.entrySet()) {
    		String entityID = entry.getKey();
    		Element entityEl = entry.getValue();
    		
//    		// do these below regardless of whether parent is set, in order to parse everything the same
			entityEl.removeAttribute("PARENT_ENTITY_ID");
			Element glossWindowItemRefsEl = doc.createElement("SEGMENT-TIER-UTTERANCE-REFS");
			entityEl.appendChild(glossWindowItemRefsEl);
    			
			var itemsList = itemElements.get(entityID);
			if (itemsList != null && itemsList.size() >= 0) {
				Element firstItem = itemsList.get(0);
				Element lastItem = itemsList.get(itemsList.size() - 1);
				var start = readLongOpt(firstItem, "S").orElse(null);
				var end = readLongOpt(lastItem, "E").orElse(null);
				
				var assumedUID = entityUtteranceMap.get(entityID);
                var assumedTID = utteranceTierMap.get(assumedUID);
                
                IntervalSet<String> tierUtterancesIntervalsSet = utteranceIntervals.get(assumedTID);
                if (tierUtterancesIntervalsSet == null)
                	tierUtterancesIntervalsSet = new IntervalSet<String>();

				var entityInterval = TimeRange.ABSOLUTE.time(start, TimeUnit.TIMESTAMP, Rounding.FLOOR)
						.to(TimeRange.ABSOLUTE.time(end, TimeUnit.TIMESTAMP, Rounding.FLOOR));

				ArrayList<String> utteranceIDs = 
						tierUtterancesIntervalsSet.testOverlap(entityInterval, Bound.INCLUSIVE).orElse(new ArrayList<String>());
				for (String utteranceID: utteranceIDs) {
					Element glossWindowItemRefEl = doc.createElement("REF");
					write(glossWindowItemRefEl, "SEGMENT-TIER-UTTERANCE_ID", utteranceID);
					glossWindowItemRefsEl.appendChild(glossWindowItemRefEl);
				}
			}
    	}
    }
    
    private void removeBlankLocationEntities() {
    	var domTrack = trackElements.get(Util.DOM_LOCATION_FIELD_ID);
    	var ndomTrack = trackElements.get(Util.NDOM_LOCATION_FIELD_ID);
    	if (domTrack == null && ndomTrack == null)
    		return;
    	
    	var locationTrackElements = List.of(domTrack, ndomTrack);

    	for (var trackEl : locationTrackElements) {
    		var locTrackID = readOpt(trackEl, "FID").orElse(null);

    		ArrayList<Element> locEntitiesToDelete = new ArrayList<Element>();

    		for (Element locEntityEl: readChildren(trackEl, "ENTITY").collect(Collectors.toCollection(ArrayList::new))) {
    			boolean allLocationRegionsEmpty = true;
    			
    			var items = readChildren(locEntityEl, "ITEM").collect(Collectors.toList());
    			for (int i = 0; i < items.size(); i++) {
    				Element itemEl = items.get(i);
    				String locationRegionID = readOpt(itemEl, "LOCATION_REGION_ID").orElse(null);
    				if (isNotBlank(locationRegionID)) {
    					if (! locationRegionID.equals("-1")) {
    						allLocationRegionsEmpty = false;
    					}
    				}
    			}
    			
    			if (allLocationRegionsEmpty) {
    				locEntitiesToDelete.add(locEntityEl);
    				
    				String refEntityID = readOpt(locEntityEl, "REF_ENTITY_ID").orElse(null);
    				Element glossEntityEl = entityElements.get(refEntityID);
    				if (glossEntityEl != null) {
    					write(glossEntityEl, "DEPENDENT_LOCATION_ID", "");
    				}
    			}
    		}
    		
    		for (Element locTrackEl: locEntitiesToDelete) {
    			trackEl.removeChild(locTrackEl);
    		}
    	}

    }
    
    private void convertTimestampsToFrames() {
    	// convert s/e values in database element
		var dbStart = readLongOpt(rootEl, "S").orElse(null);
		var dbEnd = readLongOpt(rootEl, "E").orElse(null);

		var dbStartFrames = TimeRange.ABSOLUTE.time(dbStart, TimeUnit.TIMESTAMP, Rounding.FLOOR).convert(frameTimeUnit);
		writeLong(rootEl, "S", dbStartFrames.value());

		var dbEndFrames = TimeRange.ABSOLUTE.time(dbEnd, TimeUnit.TIMESTAMP, Rounding.FLOOR).convert(frameTimeUnit);
		writeLong(rootEl, "E", dbEndFrames.value());
		
		// convert movie duration
        var mediaFilesEl = readChildOpt(rootEl, "MEDIA-FILES").orElse(null);
        	var duration = readLongOpt(mediaFilesEl, "MOVIE_DURATION").orElse(null);
        	var durationFrames = TimeRange.ABSOLUTE.time(duration, TimeUnit.TIMESTAMP, Rounding.FLOOR).convert(frameTimeUnit);
        	writeLong(mediaFilesEl, "MOVIE_DURATION", durationFrames.value());
    	
    	// convert utterance times
    	var partitionsEl = readChildOpt(glossWindowEl, "TEMPORAL-PARTITIONS").orElse(null);
    	readChildren(partitionsEl, "TEMPORAL-PARTITION").forEach(partitionEl -> {
    		var pid = readOpt(partitionEl, "ID").orElse(null);

    		readChildren(partitionEl, "SEGMENT-TIER").forEach(tierEl -> {
    			var tid = readOpt(tierEl, "ID").orElse(null);

    			var participantID = readOpt(tierEl, "PARTICIPANT_ID").orElse(null);

    			readChildren(tierEl, "SEGMENT-TIER-UTTERANCE").forEach(utteranceEl -> {
    				var uid = readOpt(utteranceEl, "ID").orElse(null);

    				var start = readLongOpt(utteranceEl, "S").orElse(null);
    				var end = readLongOpt(utteranceEl, "E").orElse(null);

    				var startFrames = TimeRange.ABSOLUTE.time(start, TimeUnit.TIMESTAMP, Rounding.FLOOR).convert(frameTimeUnit);
    				writeLong(utteranceEl, "S", startFrames.value());

    				var endFrames = TimeRange.ABSOLUTE.time(end, TimeUnit.TIMESTAMP, Rounding.FLOOR).convert(frameTimeUnit);
    				writeLong(utteranceEl, "E", endFrames.value());
    			});
    		});
    	});

    	// convert entity item times within tracks
    	var tracksEl = readChildOpt(dbEl, "TRACKS").orElse(null);
    	for (Element tEl: readChildren(tracksEl, "TRACK").collect(Collectors.toList())) {
    		var fid = readOpt(tEl, "FID").orElse(null);

    		readChildren(tEl, "ENTITY").forEach(eEl -> {
    			var eid = readOpt(eEl, "ID").orElse(null);

    			var items = readChildren(eEl, "ITEM").collect(Collectors.toList());
    			for (int i = 0; i < items.size(); i++) {
    				var iEl = items.get(i);
    				var start = readLongOpt(iEl, "S").orElse(null);
    				var end = readLongOpt(iEl, "E").orElse(null);

    				var startFrames = TimeRange.ABSOLUTE.time(start, TimeUnit.TIMESTAMP, Rounding.FLOOR).convert(frameTimeUnit);
    				writeLong(iEl, "S", startFrames.value());

    				var endFrames = TimeRange.ABSOLUTE.time(end, TimeUnit.TIMESTAMP, Rounding.FLOOR).convert(frameTimeUnit);
    				writeLong(iEl, "E", endFrames.value());
    			}
    		});

            readChildren(tEl, "NON_COMPLETE_EVENTS").forEach(eEl -> {
                var items = readChildren(eEl, "ITEM").collect(Collectors.toList());
                for (int i = 0; i < items.size(); i++) {
                    var iEl = items.get(i);
                    var start = readLongOpt(iEl, "S").orElse(null);
                    var end = readLongOpt(iEl, "E").orElse(null);

    				var startFrames = TimeRange.ABSOLUTE.time(start, TimeUnit.TIMESTAMP, Rounding.FLOOR).convert(frameTimeUnit);
    				writeLong(iEl, "S", startFrames.value());

    				var endFrames = TimeRange.ABSOLUTE.time(end, TimeUnit.TIMESTAMP, Rounding.FLOOR).convert(frameTimeUnit);
    				writeLong(iEl, "E", endFrames.value());
                }
            });

            readChildren(tEl, "ENTERED_TEXT").forEach(eEl -> {
                var items = readChildren(eEl, "ITEM").collect(Collectors.toList());
                for (int i = 0; i < items.size(); i++) {
                    var iEl = items.get(i);
                    var start = readLongOpt(iEl, "S").orElse(null);
                    var end = readLongOpt(iEl, "E").orElse(null);

    				var startFrames = TimeRange.ABSOLUTE.time(start, TimeUnit.TIMESTAMP, Rounding.FLOOR).convert(frameTimeUnit);
    				writeLong(iEl, "S", startFrames.value());

    				var endFrames = TimeRange.ABSOLUTE.time(end, TimeUnit.TIMESTAMP, Rounding.FLOOR).convert(frameTimeUnit);
    				writeLong(iEl, "E", endFrames.value());
                }
            });
    	}
    }
    
    private void convertPhaseCodes() {
    	// convert entity item phase codes, from legacy to new specific ones per field
    	var tracksEl = readChildOpt(dbEl, "TRACKS").orElse(null);
    	for (Element tEl: readChildren(tracksEl, "TRACK").collect(Collectors.toList())) {
    		var fid = readOpt(tEl, "FID").orElse(null);
            var repr = "Track " + repr(tEl, "FID");

    		readChildren(tEl, "ENTITY").forEach(eEl -> {
    			var eid = readOpt(eEl, "ID").orElse(null);
                var erepr = repr + " entity " + repr(eEl, "ID");

    			var items = readChildren(eEl, "ITEM").collect(Collectors.toList());
    			for (int i = 0; i < items.size(); i++) {
    				var iEl = items.get(i);
                    var irepr = erepr + " item";
    				
    				String txt = getCharacterDataFromElement(iEl);
    				String vid = readOpt(iEl, "VID").orElse(null);
    				String newVid = vid;
    				if (isBlank(vid)) {
    					continue;
    				}
    				
    				try {
    					int vidInt = Integer.valueOf(vid);
    					
    					if (vidInt == Constants.LEGACY_HOLD_CODE) {
    						if (txt.equals(Constants.FINAL_HOLD)) {
                        		if (fid.equals(Util.DOM_GLOSS_FIELD_ID) || fid.equals(Util.NDOM_GLOSS_FIELD_ID)) {
                        			newVid = String.valueOf(Constants.GLOSS_FINAL_HOLD_CODE);
                        		} else {
                        			// else case is all nonmanual fields
                        			newVid = String.valueOf(Constants.NONMANUAL_FINAL_HOLD_CODE);
                        		}
                        	} else if (txt.equals(Constants.INITIAL_HOLD)) {
                        		if (fid.equals(Util.DOM_GLOSS_FIELD_ID) || fid.equals(Util.NDOM_GLOSS_FIELD_ID)) {
                        			newVid = String.valueOf(Constants.GLOSS_INITIAL_HOLD_CODE);
                        		} else {
                        			// else case is all nonmanual fields
                        			newVid = String.valueOf(Constants.NONMANUAL_INITIAL_HOLD_CODE);
                        		}
                        	}
                        } else if (vidInt == Constants.LEGACY_ONSET_CODE) {
                    		if (fid.equals(Util.DOM_GLOSS_FIELD_ID) || fid.equals(Util.NDOM_GLOSS_FIELD_ID)) {
                    			newVid = String.valueOf(Constants.GLOSS_ONSET_CODE);
                    		} else {
                    			// else case is all nonmanual fields
                    			newVid = String.valueOf(Constants.NONMANUAL_ONSET_CODE);
                    		}
                        } else if (vidInt == Constants.LEGACY_OFFSET_CODE) {
                    		if (fid.equals(Util.DOM_GLOSS_FIELD_ID) || fid.equals(Util.NDOM_GLOSS_FIELD_ID)) {
                    			newVid = String.valueOf(Constants.GLOSS_OFFSET_CODE);
                    		} else {
                    			// else case is all nonmanual fields
                    			newVid = String.valueOf(Constants.NONMANUAL_OFFSET_CODE);
                    		}
                        }
    				} catch(NumberFormatException e) {
//    					e.printStackTrace();
    					continue;
    				}
    				
    				if ( ! newVid.equals(vid)) {
    					write(iEl, "VID", newVid);
    					addToModifications(0, irepr + " had phase code converted from legacy " + vid + " to correct phase code " + newVid, null);
    				}
    			}
    		});
    	}
    	
    }
    
    private void convertIsDefaultSegmentTierFormat() {
    	var partitionsEl = readChildOpt(glossWindowEl, "TEMPORAL-PARTITIONS").orElse(null);
    	readChildren(partitionsEl, "TEMPORAL-PARTITION").forEach(partitionEl -> {
            var repr = "Partition " + repr(partitionEl, "LABEL");
            
    		readChildren(partitionEl, "SEGMENT-TIER").forEach(tierEl -> {
                var trepr = repr + " segment tier " + repr(tierEl, "LABEL");
                
    			var isDefault = readOpt(tierEl, "ISDEFAULT").orElse(null);
    			var isDefaultVal = isNotBlank(isDefault) && isDefault.equals("true") ? "T" : "F";
    			
    			tierEl.removeAttribute("ISDEFAULT");
    			write(tierEl, "IS_DEFAULT", isDefaultVal);
    			
    			addToModifications(0, trepr + " had \"ISDEFAULT=" + isDefault + "\" converted to \"IS_DEFAULT=" + isDefaultVal + "\"", null);
    		});
    	});
    	
    }
}
