package edu.bu.signstream.common.textgrid;

import edu.bu.signstream.common.util.Pair;
import edu.bu.signstream.time.Interval;
import edu.bu.signstream.time.Time;
import edu.bu.signstream.time.TimeRange;
import edu.bu.signstream.time.TimeUnit;

import java.io.*;
import java.net.URI;
import java.nio.charset.Charset;

public class TextGridParser {
    private final BufferedReader reader;
    private int pos = 0;

    public static TextGrid parse(URI uri, Charset charset) {
        return new TextGridParser(uri, charset).parse();
    }

    private TextGridParser(URI uri, Charset charset) {
        // TODO: Allow selecting character encoding
        try {
            reader = new BufferedReader(new InputStreamReader(new FileInputStream(new File(uri)), charset));
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }



    public String read() {
        String line;
        do {
            try {
                line = reader.readLine();
                if (line != null)
                    line = line.trim()
                        .replace("\uFEFF", ""); // Remove UTF16 byte order mark
                pos++;
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        } while (line != null && line.isBlank());

        if (line == null)
            throw exception("Unexpected EOF");

        return line;
    }
    private RuntimeException exception(String message) { return new RuntimeException(message + " on line " + pos); }

    private void expect(String str) {
        if (!read().contains(str))
            throw exception("Expected '" + str + "'");
    }

    private String prop(String key) {
        var line = read();
        var eq = line.indexOf("=");
        if (eq == -1 || !line.substring(0, eq).contains(key))
            throw exception("Expected property '" + key + "'");
        return line.substring(eq + 1).trim();
    }
    private int integer(String key) { return Integer.parseInt(prop(key)); }
    private String string(String key) {
        var s = prop(key);
        return s.substring(1, s.length() - 1);
    }
    private Time time(String key) {
        var seconds = Double.parseDouble(prop(key));
        return TimeRange.ABSOLUTE.time(Math.round(seconds * Math.pow(10, 9)), TimeUnit.NANOSECOND);
    }



    private TextGrid parse() {
        var grid = new TextGrid();

        expect("ooTextFile"); // File type = "ooTextFile"
        expect("TextGrid"); // Object class = "TextGrid"

        grid.start = time("xmin");
        grid.end = time("xmax");

        expect("tiers");
        int n = integer("size");
        expect("item");

        for (int i = 0; i < n; i++)
            grid.addTier(parseTier());

        return grid;
    }

    private TextGrid.Tier parseTier() {
        expect("item");

        var clazz = string("class");

        if (clazz.equals("IntervalTier"))
            return parseIntervalTier();
        else if (clazz.equals("TextTier"))
            return parseTextTier();
        else
            throw exception("Unknown item class " + clazz);
    }


    private TextGrid.IntervalTier parseIntervalTier() {
        var tier = new TextGrid.IntervalTier();

        tier.name = string("name");
        if (tier.name.equals("")) tier.name = "[Unnamed]";
        tier.start = time("xmin");
        tier.end = time("xmax");

        var n = integer("intervals: size");
        for (int i = 0; i < n; i++) {
            var interval = parseInterval();

            // Intervals with text "" should be ignored.
            if (!interval.first.equals(""))
                tier.intervals.add(interval);
        }

        return tier;
    }

    private Pair<String, Interval> parseInterval() {
        expect("interval");

        var start = time("xmin");
        var end = time("xmax");
        var text = string("text");

        return Pair.of(text, start.to(end));
    }


    private TextGrid.TextTier parseTextTier() {
        var tier = new TextGrid.TextTier();

        tier.name = string("name");
        if (tier.name.equals("")) tier.name = "[Unnamed]";
        tier.start = time("xmin");
        tier.end = time("xmax");

        var n = integer("points: size");
        for (int i = 0; i < n; i++) {
            tier.points.add(parsePoint());
        }

        return tier;
    }

    private Pair<String, Time> parsePoint() {
        expect("points");

        var point = time("number");
        var mark = string("mark");

        return Pair.of(mark, point);
    }
}
