package edu.bu.signstream.grepresentation.fields;

import edu.bu.signstream.time.TimeUnit;

import java.awt.*;
import java.util.stream.Collectors;

public class ImageEventsEntityPainter {
    private static final int W_IMAGE = 40;
    private static final int H_IMAGE = 24;

    private static final int W_PAD = 0;
    private static final int W_RECT = 3;

    private static final int DY_IMAGE = H_IMAGE / 2;
    private static final int DY_CAP = 4;

    public static <T extends EventsEntity & ImageEventsEntity> void paint(Graphics2D g, int y, int h, T entity) {
        var SSImages = entity.images();
        int n = SSImages.size();
        
        // if 2 hs, but they're both the same (according to id), then treat it as if it's just 1 hs
        if (n == 2) {
        		var ssimage1 = SSImages.get(0);
        		var ssimage2 = SSImages.get(1);
        		
        		if (ssimage1 != null && ssimage2 != null) {
        			var ssimage1id = ssimage1.getID();
        			var ssimage2id = ssimage2.getID();
        			if (ssimage1id.equals(ssimage2id)) {
        				n = 1;
        				SSImages.remove(1);
        			}
        		}
        }
        
        	var images = SSImages.stream().map(i -> i.getImage()).collect(Collectors.toList());

        var interval = entity.interval();
        var x0 = interval.t0.as(TimeUnit.COORDINATE);
        var x1 = interval.t1.as(TimeUnit.COORDINATE);
        var w = x1 - x0;

        g.setColor(entity.color());

        // Draw the baseline and start/end markers
        var base = y + h/2;
        g.drawLine((int)Math.round(x0), base, (int)Math.round(x1), base);
        g.drawLine((int)Math.round(x0), base + DY_CAP, (int)Math.round(x0), base - DY_CAP);
        g.drawLine((int)Math.round(x1), base + DY_CAP, (int)Math.round(x1), base - DY_CAP);

        // Special rules for entities with only 1 image.
        if (n == 1 && x0 == x1) {
            // For zero length entities, draw a single centered image no matter the size.
            image(g, images.get(0), x0 + w/2 - W_IMAGE/2.0, base);
            return;

        } else if (n == 1 && w < W_IMAGE) {
            // If the image doesn't fit, draw a single centered rectangle.
            rect(g, images.get(0), x0 + w / 2, base);
            return;

        } else if (n == 1 && w < W_IMAGE*2 + W_PAD) {
            // If two images don't fit, draw a single centered image.
            image(g, images.get(0), x0 + w/2 - W_IMAGE/2.0, base);
            return;

        } else if (n == 1) {
            // If two can fit, duplicate the image and proceed with the normal logic.
            n++;
            images.add(images.get(0));
        }

        if (w > n*W_IMAGE + (n - 1)*W_PAD) {
            // If all images fit, draw first, last, and rest evenly spaced.
            image(g, images.get(0), x0, base);
            var dx = (w - W_IMAGE) / (n - 1);
            for (int i = 1; i < n - 1; i++)
                image(g, images.get(i), x0 + i*dx, base);
            image(g, images.get(n - 1), x1 - W_IMAGE, base);

        } else if (w > 2*W_IMAGE + (n - 1)*W_PAD + (n - 2)*W_RECT) {
            // If two images fit, draw first, last, and rest rectangles.
            image(g, images.get(0), x0, base);
            var dx = (w - 2*W_IMAGE) / (n - 1);
            for (int i = 1; i < n - 1; i++)
                rect(g, images.get(i), x0 + W_IMAGE + i*dx, base);
            image(g, images.get(n - 1), x1 - W_IMAGE, base);

        } else if (w > W_IMAGE + (n - 1)*W_PAD + (n - 1)*W_RECT) {
            // If only one image fits, draw first and rest rectangles.
            image(g, images.get(0), x0, base);
            var dx = (w - W_IMAGE) / (n - 1);
            for (int i = 1; i < n; i++)
                rect(g, images.get(i), x0 + W_IMAGE - W_RECT + i*dx, base);

        } else if (w > n*W_RECT + (n - 1)*W_PAD) {
            // If no images fit, draw rectangles if they fit.
            rect(g, images.get(0), x0 + 1, base);
            var dx = w / (n - 1);
            for (int i = 1; i < n - 1; i++)
                rect(g, images.get(i), x0 + i*dx, base);
            rect(g, images.get(n - 1), x1 - 3, base);
        }
    }

    private static void image(Graphics2D g, Image i, double x, int y) {
        if (i == null) return;

        var X = (int)Math.round(x);
        g.drawImage(i, X, y - DY_IMAGE, Color.WHITE, null);
        g.drawRect(X, y - DY_IMAGE, i.getWidth(null), i.getHeight(null));
    }

    private static void rect(Graphics2D g, Image i, double x, int y) {
        if (i == null) return;

        g.drawRect((int)Math.round(x), y - DY_IMAGE, W_RECT - 1, H_IMAGE);
    }
}
