package edu.bu.signstream.media.gstreamer;


import java.awt.*;
import java.net.URI;

import javax.swing.*;

import edu.bu.signstream.time.Duration;
import edu.bu.signstream.time.TimeRange;
import edu.bu.signstream.time.TimeUnit;
import edu.bu.signstream.media.Video;
import edu.bu.signstream.media.ui.VideoView;
import edu.bu.signstream.ui.frame.SSFrame;
import org.freedesktop.gstreamer.*;
import org.freedesktop.gstreamer.elements.AppSink;
import org.freedesktop.gstreamer.query.PositionQuery;

public class GstVideo implements GstMedia, Video {
    private static final String FORMAT = "video/x-raw";
    private static final String ID_PREFIX = "Video";

    private static int ID = 0;
    private final String id;
    private int order;

    private final URI uri;

    private int width;
    private int height;
    private double framerate;
    private Duration duration;

    private TimeRange range;

    private int timestep;
    private int timebase;

    private final GstMediaPlayer player;

    private final Element decode;
    private final Element convert;
    private final Element sink;

    private final GstVideoImage image;

    private VideoView view;

    public GstVideo(URI uri, int order, GstMediaPlayer player) {
        this.player = player;
        this.uri = uri;
        this.order = order;
        this.id = ID_PREFIX + ID;

        System.out.println("Probing video " + uri.toString());
        var caps = GstProbe.probe(uri, FORMAT);
        System.out.println("Probing finished!");

        var fr = (Fraction)caps.get("framerate");
        timestep = fr.getDenominator();
        timebase = fr.getNumerator();
        framerate = (double) timebase / (double) timestep;
        var unit = new TimeUnit(id, "FRAMES", TimeUnit.Type.FRAME, framerate);

        duration = ((Duration)caps.get("duration")).convert(unit);
        range = TimeRange.ABSOLUTE.subrange(id, unit, Duration.ZERO, duration, 1);

        width = (int)caps.get("width");
        height = (int)caps.get("height");

        // Set up the elements for the real pipeline
        decode = ElementFactory.make("uridecodebin", "decode" + id);
        decode.set("uri", uri.toString());
        decode.connect(new Element.PAD_ADDED() {
            @Override
            public void padAdded(Element e, Pad pad) {
                System.out.println(uri + ": " + pad.getName() + " [" + pad.getTypeName() + "]");

                if (pad.getName().equals("src_0")) {
                    try {
                        pad.link(convert.getStaticPad("sink"));
                    } catch (PadLinkException ex) {
                        ex.printStackTrace();
                    }
                }
            }
        });
        convert = ElementFactory.make("videoconvert", "convert" + id);

        var appsink = new AppSink("sink" + id);
        image = new GstVideoImage(appsink);
        image.setPreferredSize(new Dimension(width, height));
        sink = appsink;

        player.pipeline.addMany(decode, convert, sink);
        convert.link(sink);

        System.out.println("Done loading video " + id);
        ID++;
    }

    String id() {
        return id;
    }
    
    @Override
	public int order() {
    	return order;
    }

    @Override
    public URI uri() {
        return uri;
    }

    @Override
    public int width() {
        return width;
    }

    @Override
    public int height() {
        return height;
    }

    @Override
    public double framerate() {
        return framerate;
    }

    @Override
    public Duration duration() {
        return duration;
    }

    @Override
    public TimeRange range() {
        return range;
    }

    @Override
    public void range(TimeRange range) {
        this.range = range;
    }

    @Override
    public int timestep() {
        return timestep;
    }

    @Override
    public int timebase() {
        return timebase;
    }

    @Override
    public JComponent image() {
        return image;
    }

    @Override
    public SSFrame view() { return view; }

    @Override
    public void display() {
        view = new VideoView(player, this, image, this::hide);
    }

    @Override
    public void hide() {
        if (view != null)
            view.dispose();
        view = null;
    }

    @Override
    public boolean displayed() {
        return view != null;
    }

    @Override
    public long pos() {
        if (!Thread.currentThread().getName().contains("gstreamer service thread"))
            throw new AssertionError("Call to GstVideo.pos() from outside of GStreamer service thread");

        var q = new PositionQuery(Format.TIME);
        sink.query(q);
        return q.getPosition();
    }

    @Override
    public Element sink() {
        return sink;
    }

    @Override
    public void close() {
        if (!Thread.currentThread().getName().contains("gstreamer service thread"))
            throw new AssertionError("Call to GstVideo.close() from outside of GStreamer service thread");

        player.pipeline.remove(decode);
        player.pipeline.remove(convert);
        player.pipeline.remove(sink);
    }
}
