package edu.bu.signstream.media.ui;

import edu.bu.signstream.common.util.Util;
import edu.bu.signstream.grepresentation.reguestsHandler.SS3Singleton;
import edu.bu.signstream.time.Duration;
import edu.bu.signstream.time.Time;
import edu.bu.signstream.time.TimeRange;
import edu.bu.signstream.time.TimeUnit;
import edu.bu.signstream.media.Media;
import edu.bu.signstream.media.MediaPlayer;
import edu.bu.signstream.media.ui.icon.*;
import edu.bu.signstream.ui.SS3SignStreamApplication;
import edu.bu.signstream.ui.frame.Blocker;
import edu.bu.signstream.ui.frame.SSFrame;

import javax.swing.*;
import java.awt.*;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.function.Consumer;

public abstract class MediaView extends SSFrame {
    private final ToggleButton play = new ToggleButton(b -> b ? new PauseIcon(true) : new PlayIcon(true), false);
    private final JButton slower = new ToggleButton(SpeedSlowerIcon::new);
    private final JButton faster = new ToggleButton(SpeedFasterIcon::new);

    private final ToggleButton forward1 = new ToggleButton(GoFrameForward1Icon::new);
    private final ToggleButton back1 = new ToggleButton(GoFrameBack1Icon::new);
    private final ToggleButton forward2 = new ToggleButton(GoFrameForward2Icon::new);
    private final ToggleButton back2 = new ToggleButton(GoFrameBack2Icon::new);
    private final ToggleButton forward3 = new ToggleButton(GoFrameForward3Icon::new);
    private final ToggleButton back3 = new ToggleButton(GoFrameBack3Icon::new);
    private final ToggleButton start = new ToggleButton(GoFirstFrameIcon::new);
    private final ToggleButton end = new ToggleButton(GoLastFrameIcon::new);

    protected static final int SLIDER_MAX = 1000000;
    protected final JSlider slider = new JSlider(JSlider.HORIZONTAL, SLIDER_MAX);

    private final JLabel rate = new JLabel();
    private final JLabel time = new JLabel("0 s");

    private final Consumer<Time> timeListener;
    private final Consumer<Double> rateListener;
    private final Consumer<Boolean> stateListener;
    protected boolean updating = false;

    protected MediaPlayer player;
    protected Media media;

    public MediaView(MediaPlayer player, Media media, Runnable onClose) {
        this.player = player;
        this.media = media;

        timeListener = t -> SwingUtilities.invokeLater(() -> {
            updating = true;
            var fr = t.in(media.range()).duration()
                .divide(media.range().duration());
            fr = Math.max(0, Math.min(1, fr));
            timeChanged(t, fr);
            updating = false;
        });
        rateListener = r -> SwingUtilities.invokeLater(() -> rateChanged(r));
        stateListener = s -> SwingUtilities.invokeLater(() -> stateChanged(s));

        player.addTimeListener(timeListener);
        player.addRateListener(rateListener);
        player.addStateListener(stateListener);

        addWindowListener(new WindowAdapter() {
            @Override
            public void windowClosing(WindowEvent e) {
                super.windowClosing(e);
                player.removeTimeListener(timeListener);
                player.removeRateListener(rateListener);
                player.removeStateListener(stateListener);
                onClose.run();
            }
        });
    }

    protected void timeChanged(Time t, double fr) {
        slider.setValue((int)Math.round(fr * SLIDER_MAX));
        time.setText(String.format("%,.2fs", t.absolute().as(TimeUnit.SECOND)));
    }

    protected void rateChanged(double r) {
        rate.setText(r == 1 ? "" : "X" + r);
    }

    protected void stateChanged(boolean playing) {
        play.set(playing);

        start.set(!playing);
        back3.set(!playing);
        back2.set(!playing);
        back1.set(!playing);
        forward1.set(!playing);
        forward2.set(!playing);
        forward3.set(!playing);
        end.set(!playing);
    }

    protected abstract JComponent view();
    protected abstract JPanel controls(JPanel panel);

    protected Dimension minSize() {
        return new Dimension(300, 200);
    }

    protected abstract TimeRange defaultTimeRange();

    protected void build() {
        var playPauseGroup = new JPanel(new FlowLayout(FlowLayout.CENTER, 0, 0));
        playPauseGroup.setAlignmentX(Component.LEFT_ALIGNMENT);
        playPauseGroup.setBackground(Color.BLACK);
        slower.addActionListener(l -> { if (player.rate() >= 0.5) player.rate(player.rate() / 2); });
        slower.setToolTipText("Slower");
        playPauseGroup.add(slower);
        play.addActionListener(l -> { if (play.state()) player.pause(); else player.play(); });
        play.setToolTipText("Play");
        playPauseGroup.add(play);
        faster.addActionListener(l -> { if (player.rate() <= 4.0) player.rate(player.rate() * 2); });
        faster.setToolTipText("Faster");
        playPauseGroup.add(faster);

        var controlsGroup = new JPanel(new FlowLayout(FlowLayout.CENTER, 0, 0));
        controlsGroup.setAlignmentX(Component.RIGHT_ALIGNMENT);
        controlsGroup.setBackground(Color.BLACK);
        controlsGroup = controls(controlsGroup);

        var stepGroup = new JPanel(new FlowLayout(FlowLayout.CENTER, 0, 0));
        stepGroup.setBackground(Color.BLACK);
        start.addActionListener(l -> { if (start.state()) player.seek(media.range().start()); });
        start.setToolTipText("Start");
        stepGroup.add(start);
        back3.addActionListener(l -> {
            if (back3.state())
                player.seek(player.time().subtract(Duration.ms(1000))
                    .in(Util.getCurrentRange(defaultTimeRange())));
            }
        );
        back3.setToolTipText("1 Second Backward");
        stepGroup.add(back3);
        back2.addActionListener(l -> {
            if (back2.state())
                player.seek(player.time().subtract(Duration.ms(500))
                    .in(Util.getCurrentRange(defaultTimeRange())));
            }
        );
        back2.setToolTipText("0.5 Seconds Backward");
        stepGroup.add(back2);
        back1.addActionListener(l -> {
            if (back1.state()) {
                var range = defaultTimeRange();
                player.seek(player.time().subtract(new Duration(range.step(), range.unit()))
                    .in(range));
            }
        });
        back1.setToolTipText("1 Frame Backward");
        stepGroup.add(back1);
        forward1.addActionListener(l -> {
            if (forward1.state()) {
                var range = defaultTimeRange();
                player.seek(player.time().add(new Duration(range.step(), range.unit()))
                    .in(range));
            }
        });
        forward1.setToolTipText("1 Frame Forward");
        stepGroup.add(forward1);
        forward2.addActionListener(l -> {
            if (forward2.state())
                player.seek(player.time().add(Duration.ms(500))
                    .in(Util.getCurrentRange(defaultTimeRange())));
            }
        );
        forward2.setToolTipText("0.5 Seconds Forward");
        stepGroup.add(forward2);
        forward3.addActionListener(l -> {
            if (forward3.state())
                player.seek(player.time().add(Duration.ms(1000))
                    .in(Util.getCurrentRange(defaultTimeRange())));
            }
        );
        forward3.setToolTipText("1 Second Forward");
        stepGroup.add(forward3);
        end.addActionListener(l -> { if (end.state()) player.seek(media.range().end()); });
        end.setToolTipText("End");
        stepGroup.add(end);

        var strip = new JPanel();
        strip.setLayout(new BoxLayout(strip, BoxLayout.X_AXIS));
        strip.setBackground(Color.BLACK);
        strip.setMinimumSize(new Dimension(300, 32));
        strip.setPreferredSize(new Dimension(300, 32));
        strip.setMaximumSize(new Dimension(Integer.MAX_VALUE, 32));

        strip.add(playPauseGroup);
        strip.add(Box.createHorizontalGlue());
        strip.add(stepGroup);
        if (controlsGroup != null) {
            strip.add(Box.createHorizontalGlue());
            strip.add(controlsGroup);
        }

        var container = new JPanel();
        container.setLayout(new BoxLayout(container, BoxLayout.Y_AXIS));
        container.add(view());
        container.add(strip);
        container.add(slider);
        slider.addChangeListener(c -> {
            if (!updating) {
                var fr = (double) slider.getValue() / SLIDER_MAX;
                var t = media.range().zero()
                    .add(media.range().duration().multiply(fr));

                player.pause();
                player.seek(t);
            }
        });
        slider.setFocusable(false);
        updating = true;
        slider.setValue(0);
        updating = false;

        child().add(container);

        var glass = (JPanel)getGlassPane();
        glass.setLayout(new BoxLayout(glass, BoxLayout.Y_AXIS));
        time.setFont(new Font(rate.getFont().getName(), Font.PLAIN, 22));
        time.setForeground(Color.GREEN);
        glass.add(time);
        rate.setFont(new Font(rate.getFont().getName(), Font.PLAIN, 22));
        rate.setForeground(Color.WHITE);
        glass.add(rate);
        glass.setVisible(true);

        setTitle(media.uri().toString());
        setMinimumSize(minSize());

        addKeyListener(new KeyAdapter() {
            @Override
            public void keyPressed(KeyEvent e) {
                var key = e.getKeyCode();

                var range = media.range();
                var step = Util.getStepInRange(range);

                switch (key) {
                    case KeyEvent.VK_LEFT:
                        if (!blocked())
                            player.seek(player.time().subtract(step).in(range));
                        break;

                    case KeyEvent.VK_RIGHT:
                        if (!blocked())
                            player.seek(player.time().add(step).in(range));
                        break;

                    case KeyEvent.VK_SPACE:
                        if (!blocked())
                            play.doClick();
                        break;

                    case KeyEvent.VK_S:
                        if (e.isMetaDown())
                            SS3Singleton.getSignStreamApplication().getJMenuBar().getMenu(0)
                                .getItem(4)
                                .doClick();
                        break;

                    case KeyEvent.VK_N:
                        if (e.isMetaDown())
                            SS3Singleton.getSignStreamApplication().getJMenuBar().getMenu(0)
                                .getItem(0)
                                .doClick();
                        break;

                    case KeyEvent.VK_O:
                        if (e.isMetaDown())
                            SS3Singleton.getSignStreamApplication().getJMenuBar().getMenu(0)
                                .getItem(1)
                                .doClick();
                        break;

                    case KeyEvent.VK_W:
                        if (e.isMetaDown())
                            SS3Singleton.getSignStreamApplication().getJMenuBar().getMenu(0)
                                .getItem(3)
                                .doClick();
                        break;

                    default:
                        break;
                }
            }
        });

        pack();
        setVisible(true);
    }

    @Override
    public void dispose() {
        player.removeTimeListener(timeListener);
        player.removeRateListener(rateListener);
        player.removeStateListener(stateListener);

        super.dispose();
    }
}
