package edu.bu.signstream.media.ui.painter;

import edu.bu.signstream.common.SSMath;
import edu.bu.signstream.common.util.Pair;
import edu.bu.signstream.time.Duration;
import edu.bu.signstream.time.TimeRange;
import edu.bu.signstream.time.TimeUnit;
import edu.bu.signstream.grepresentation.reguestsHandler.SS3Singleton;
import edu.bu.signstream.media.data.Samples;
import edu.bu.signstream.media.ui.color.ColorMap;

import javax.swing.*;
import java.awt.*;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

public abstract class SamplesPainter {
    protected final Samples samples;
    protected final ColorMap colors;
    protected final TimeRange range;

    protected double pageSamples;
    protected double pages;

    protected double fr;
    protected int page;

    private static final int CACHE_CAPACITY = 100;
    private final LinkedHashMap<Pair<Integer, Dimension>, Image> cache = new LinkedHashMap<>(CACHE_CAPACITY, 2.0f, true) {
        @Override
        protected boolean removeEldestEntry(Map.Entry<Pair<Integer, Dimension>, Image> eldest) {
            return size() > CACHE_CAPACITY;
        }
    };

    protected boolean cursor = false;
    protected int mx = 0;
    protected int my = 0;

    public SamplesPainter(Samples samples, ColorMap colors, TimeRange range, Duration page) {
        this.samples = samples;
        this.colors = colors;
        this.range = range;
        page(page);
    }

    public void page(Duration duration) {
        cache.clear();
        pageSamples = (int)(samples.rate * duration.as(TimeUnit.SECOND));
        pages = (double)samples.length / pageSamples;
        page = (int)(fr * pages);
    }

    // Update the current time (and the current page)
    public void seek(double fr) {
        this.fr = fr;
        page = (int)(fr * pages);
    }

    public void addListeners(JComponent component) {
        component.addComponentListener(new ComponentAdapter() {
            @Override
            public void componentResized(ComponentEvent e) {
                cache.clear();
            }
        });

        var l = new MouseAdapter() {
            @Override
            public void mousePressed(MouseEvent e) {
                var w = component.getSize().getWidth();
                var fr = SSMath.ilerp(0, pages, page + SSMath.ilerp(0, w, e.getX()));
                var t = range.zero().add(range.duration().multiply(fr));

                var player = SS3Singleton.getMediaPlayer();
                player.pause();
                player.seek(t);
            }
            @Override public void mouseDragged(MouseEvent e) { mousePressed(e); }

            @Override
            public void mouseEntered(MouseEvent e) {
                cursor = true;
                mx = e.getX();
                my = e.getY();
                component.repaint();
            }

            @Override
            public void mouseExited(MouseEvent e) {
                cursor = false;
                component.repaint();
            }

            @Override
            public void mouseMoved(MouseEvent e) {
                mx = e.getX();
                my = e.getY();
                component.repaint();
            }
        };
        component.addMouseListener(l);
        component.addMouseMotionListener(l);
    }

    protected abstract void paint(Graphics g, int w, int h, int s0, int s1);

    protected List<String> text(int w, int h, int mx, int my) { return new ArrayList<>(); }
    public List<String> text(JComponent parent) {
        var sz = parent.getSize();
        return text(sz.width, sz.height, mx, my);
    }

    public void paint(JComponent parent, Graphics g) {
        var sz = parent.getSize();

        var image = cache.get(Pair.of(page, sz));
        if (image == null) {
            image = new BufferedImage(sz.width, sz.height, BufferedImage.TYPE_4BYTE_ABGR);

            // Calculate start and end sample for the current page
            int s0 = (int) Math.round(page * pageSamples);
            int s1 = (int) Math.round((page + 1) * pageSamples);
            try {
                paint(image.getGraphics(), sz.width, sz.height, s0, s1);
            } catch (Exception e) {
                e.printStackTrace();
            }

            cache.put(Pair.of(page, sz), image);
        }

        g.drawImage(image, 0, 0, null);
    }

    public void paintBar(JComponent parent, Graphics g) {
        var sz = parent.getSize();

        int bar = (int)Math.round(sz.width * SSMath.ilerp(page, page + 1, fr * pages));
        g.setColor(Color.GREEN);
        g.drawLine(bar, 0, bar, sz.height);
    }

    public void paintCursor(JComponent parent, Graphics g, Color color, List<String> text) {
        if (!cursor) return;

        var sz = parent.getSize();

        // Crosshair
        g.setColor(color);
        g.drawLine(mx, 0, mx, sz.height);
        g.drawLine(0, my, sz.width, my);

        // Crosshair text
        var fr = SSMath.ilerp(0, pages, page + SSMath.ilerp(0, sz.width, mx));
        var t = range.zero().add(range.duration().multiply(fr))
            .absolute().as(TimeUnit.SECOND);
        text.add("Time: " + (Math.round(t * 100.0) / 100.0) + " s");

//        var s = text.stream().reduce(String::concat).orElse("");
        var bounds = text.stream()
            .map(s -> g.getFont()
                  .createGlyphVector(g.getFontMetrics().getFontRenderContext(), s)
                  .getVisualBounds()
                  .getBounds2D())
            .map(r -> Pair.of((int)Math.round(r.getWidth()), (int)Math.round(r.getHeight())))
            .reduce((a, b) -> Pair.of(Math.max(a.first, b.first), Math.max(a.second, b.second)))
            .orElse(Pair.of(0, 0));

        var tx = mx + (mx > sz.width / 2 ? -20 - bounds.first : 20);
        var ty = my + (my > sz.height / 2 ? -20 * text.size() : 20);

        for (int i = 0; i < text.size(); i++)
            g.drawString(text.get(i), tx, ty + (i * 20));
    }
    public void paintCursor(JComponent parent, Graphics g, Color color) { paintCursor(parent, g, color, new ArrayList<>()); }
}
