package edu.bu.signstream.time;

public class Duration implements Comparable<Duration> {
    private final long value;
    private final TimeUnit unit;

    public Duration(long duration, TimeUnit unit) {
        this.value = duration;
        this.unit = unit;
    }

    public static final Duration ZERO = new Duration(0, TimeUnit.NANOSECOND);

    public static Duration sec(double seconds) {
        return ns(TimeUnit.SECOND.convert(seconds, TimeUnit.NANOSECOND, Rounding.NEAREST));
    }
    public static Duration ms(double milliseconds) {
        return ns(TimeUnit.MILLISECOND.convert(milliseconds, TimeUnit.NANOSECOND, Rounding.NEAREST));
    }
    public static Duration us(double microseconds) {
        return ns(TimeUnit.MICROSECOND.convert(microseconds, TimeUnit.NANOSECOND, Rounding.NEAREST));
    }
    public static Duration ns(long nanoseconds) { return new Duration(nanoseconds, TimeUnit.NANOSECOND); }

    @Deprecated
    public static Duration ts(long timestamp) { return new Duration(timestamp, TimeUnit.TIMESTAMP); }


    public long value() { return value; }
    public TimeUnit unit() { return unit; }

    /**
     * Converts the Duration into a Time relative to the start of a range.
     */
    public Time in(TimeRange range) {
        return range.zero().add(this);
    }

    /**
     * Converts the Duration into an integer number of the target unit.
     */
    public long as(TimeUnit unit, Rounding r) {
        return this.unit.convert(value, unit, r);
    }
    /**
     * Converts the Duration into the target unit.
     */
    public double as(TimeUnit unit) { return this.unit.convert(value, unit); }

    /**
     *
     * @return
     */

    public long absolute() { return as(TimeUnit.NANOSECOND, Rounding.NEAREST); }

    /**
     * Converts the relative time into an integer number of the target unit,
     * and returns a new Duration with the new value and unit.
     */
    public Duration convert(TimeUnit unit, Rounding r) {
        return new Duration(as(unit, r), unit);
    }
    public Duration convert(TimeUnit unit) {
        return new Duration(as(unit, Rounding.NEAREST), unit);
    }


    public Duration add(Duration o, Rounding r) {
        return new Duration(value + o.as(unit, r), unit);
    }
    public Duration add(Duration o) {
        return new Duration(value + o.as(unit, Rounding.NEAREST), unit);
    }

    public Duration subtract(Duration o, Rounding r) {
        return new Duration(value - o.as(unit, r), unit);
    }
    public Duration subtract(Duration o) {
        return new Duration(value - o.as(unit, Rounding.NEAREST), unit);
    }

    public Duration multiply(double factor, Rounding r) {
        return new Duration(r.round(value * factor), unit);
    }
    public Duration multiply(double factor) { return multiply(factor, Rounding.NEAREST); }

    public Duration divide(double factor, Rounding r) {
        return new Duration(r.round(value / factor), unit);
    }
    public Duration divide(double factor) { return divide(factor, Rounding.NEAREST); }

    public double divide(Duration o, Rounding r) {
        return (double)value / o.as(unit, r);
    }
    public double divide(Duration o) {
        return (double)absolute() / o.absolute();
    }

    @Override
    public int compareTo(Duration o) {
        return Long.compare(absolute(), o.as(TimeUnit.NANOSECOND, Rounding.NEAREST)); }

    @Override
    public boolean equals(Object o) { return o instanceof Duration && compareTo((Duration) o) == 0; }

    public boolean greater(Duration o, Bound b) {
        return b == Bound.INCLUSIVE ? compareTo(o) >= 0 : compareTo(o) > 0;
    }
    public boolean greaterThan(Duration o) { return greater(o, Bound.EXCLUSIVE); }
    public boolean greaterThanOrEq(Duration o) { return greater(o, Bound.INCLUSIVE); }

    public boolean less(Duration o, Bound b) {
        return b == Bound.INCLUSIVE ? compareTo(o) <= 0 : compareTo(o) < 0;
    }
    public boolean lessThan(Duration o) { return less(o, Bound.EXCLUSIVE); }
    public boolean lessThanOrEq(Duration o) { return less(o, Bound.INCLUSIVE); }

    public Duration max(Duration o) { return greaterThanOrEq(o) ? this : o; }
    public Duration min(Duration o) { return lessThanOrEq(o) ? this : o; }

    @Override
    public String toString() {
        return "Duration{" + value + " " + unit.suffix() + "}";
    }
}
