package edu.bu.signstream.ui.panels;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.stream.Collectors;

import javax.swing.*;

import edu.bu.signstream.common.util.Util;
import edu.bu.signstream.common.util.vo.ss3.db.SS3GlossWindowSegment;
import edu.bu.signstream.common.util.vo.ss3.db.SS3TemporalPartition;
import edu.bu.signstream.grepresentation.view.UtteranceView;
import org.jdesktop.swingx.renderer.DefaultListRenderer;

public class SegmentSelectionPanel extends JPanel {

	private static final long serialVersionUID = 1L;
	
	private JLabel selectionLabel = null;

	private JComboBox<String> partitionComboBox = new JComboBox<String>();
	private String currentPartitionString = null;
	
	private JComboBox<String> segmentComboBox = new JComboBox<String>();
	private String currentSegmentString= null;
	
	private DefaultComboBoxModel[] models;
	private DefaultListRenderer[] renderers;
	private String[] defaultSegmentTiers;
	
	private JButton enterButton = null;
	
	private JButton cancelButton = null; 
	
	private SegmentSelectionDialog segmentSelectionDialog = null;
	private UtteranceView utteranceView = null;
	
	public SegmentSelectionPanel(SegmentSelectionDialog segmentSelectionDialog, 
			UtteranceView utteranceView) {
		this.segmentSelectionDialog = segmentSelectionDialog;
		this.utteranceView = utteranceView;
		
		selectionLabel = new JLabel(
				"<html><body> <br> Please select the Temporal Partition and Segment Tier you wish to view. <br> </body></html>");
		
		ArrayList<SS3TemporalPartition> temporalPartitions = 
				utteranceView.getSignStreamSegmentPanel().getSS3Database().getSS3GlossWindow().getTemporalPartitions();
		int tpSize = temporalPartitions.size();
		
		// combobox model array
		models = new DefaultComboBoxModel[tpSize];
		renderers = new DefaultListRenderer[tpSize];
		defaultSegmentTiers = new String[tpSize];
		
		for (int i = 0; i < temporalPartitions.size(); i++) {
			SS3TemporalPartition temporalPartition = temporalPartitions.get(i);
			String partitionLabel = temporalPartition.getLabel();
			partitionComboBox.addItem(partitionLabel);

			ArrayList<SS3GlossWindowSegment> temporalPartitionSegments = temporalPartition.getTemporalPartitionSegments();
			int stSize = temporalPartitionSegments.size();	
			
			// each TP selection is associated with a different model of ST strings
			String[] segmentTierLabels = new String[stSize];

			// find the default segment to set as selected when TP is switched
			SS3GlossWindowSegment defaultSegment = temporalPartition.getDefaultSegment();
			String defaultSegmentString = null;

            var selectedSegments = utteranceView.getSegments().stream()
                .map(s -> s.getParentField()
                    .getSS3GlossWindowSegment())
                .collect(Collectors.toSet());

			for (int k = 0; k < stSize; k++) {
				SS3GlossWindowSegment temporalPartitionSegment = temporalPartitionSegments.get(k);
				String segmentLabel = temporalPartitionSegment.getLabel();
				String participantLabel = temporalPartitionSegment.getParticipantLabel();
				String combinedLabel = participantLabel + Util.SEGMENT_TIER_SEPARATOR + segmentLabel;

				if (combinedLabel.indexOf(Util.SEGMENT_TIER_SEPARATOR) == 0)
					combinedLabel = combinedLabel.substring(1);
				
				segmentTierLabels[k] = combinedLabel;
			}

            if (!selectedSegments.contains(defaultSegment)) {
                defaultSegmentString = defaultSegment.getParticipantLabel() + Util.SEGMENT_TIER_SEPARATOR + defaultSegment.getLabel();
            } else {
                defaultSegmentString = temporalPartitionSegments.stream()
                    .filter(s -> !selectedSegments.contains(s))
                    .map(s -> s.getParticipantLabel() + Util.SEGMENT_TIER_SEPARATOR + s.getLabel())
                    .findFirst()
                    .orElse(null);
            }

            models[i] = new DefaultComboBoxModel<String>(segmentTierLabels);
            renderers[i] = new DefaultListRenderer() {
                @Override
                public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected, boolean cellHasFocus) {
                    var c = super.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);

                    if (index != -1) {
                        var segment = temporalPartitionSegments.get(index);
                        if (selectedSegments.contains(segment)) {
                            c.setForeground(Color.GRAY);
                            c.setBackground(Color.LIGHT_GRAY);
                        } else if (isSelected) {
                            c.setBackground(Color.BLUE);
                        }
                    }

                    return c;
                }
            };
			defaultSegmentTiers[i] = defaultSegmentString;
		}
		
		// guaranteed to have at least one index (i.e. at least one TP in the collection)
		segmentComboBox.setModel(models[0]);
		segmentComboBox.setRenderer(renderers[0]);
		segmentComboBox.setSelectedItem(defaultSegmentTiers[0]);

        segmentComboBox.addActionListener(l -> {
            if (segmentComboBox.getSelectedIndex() == -1)
                return;

            var segments = temporalPartitions.get(partitionComboBox.getSelectedIndex())
                .getTemporalPartitionSegments();
            var segment = segments.get(segmentComboBox.getSelectedIndex());

            var selectedSegments = utteranceView.getSegments().stream()
                .map(s -> s.getParentField()
                    .getSS3GlossWindowSegment())
                .collect(Collectors.toSet());

            if (selectedSegments.contains(segment)) {
                segmentComboBox.setSelectedItem(defaultSegmentTiers[partitionComboBox.getSelectedIndex()]);
            }
        });
		
		partitionComboBox.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				int selectedIndex = partitionComboBox.getSelectedIndex();
				
				segmentComboBox.setModel(models[selectedIndex]);
				segmentComboBox.setRenderer(renderers[selectedIndex]);
				segmentComboBox.setSelectedItem(defaultSegmentTiers[selectedIndex]);
			}
		});	
		
		enterButton = new JButton("Confirm Selection");
		enterButton.addActionListener(new ConfirmSelectionListener(this));
		
		cancelButton = new JButton("Cancel");
		cancelButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				getParentDialog().dispose();
			}
		});
		
		setLayout(new GridLayout(0, 1));
		add(selectionLabel);
		add(partitionComboBox);
		add(segmentComboBox);
		add(enterButton);
		add(cancelButton);
	}
	
	public SegmentSelectionDialog getParentDialog() {
		return segmentSelectionDialog;
	}
	
	public UtteranceView getUtteranceView() {
		return utteranceView;
	}
	
	public String getSelectedPartitionString() {
		return (String) partitionComboBox.getSelectedItem();
	}
	
	public String getSelectedSegmentString() {
		return (String) segmentComboBox.getSelectedItem();
	}
}

class ConfirmSelectionListener implements ActionListener {
	SegmentSelectionPanel segmentSelectionPanel = null;
	
	public ConfirmSelectionListener(SegmentSelectionPanel segmentSelectionPanel) {
		this.segmentSelectionPanel = segmentSelectionPanel;
	}
	
	@Override
	public void actionPerformed(ActionEvent e) {
	    if (segmentSelectionPanel.getSelectedSegmentString() == null) {
	        JOptionPane.showMessageDialog(segmentSelectionPanel,
                "Please select a valid segment tier that is not already displayed.",
                "Entry Error", JOptionPane.WARNING_MESSAGE);
	        return;
        }

		String temporalPartitionLabel = segmentSelectionPanel.getSelectedPartitionString();
		String segmentTierWithParticipantLabel = segmentSelectionPanel.getSelectedSegmentString();
		String segmentTierLabel = segmentTierWithParticipantLabel.
				substring(segmentTierWithParticipantLabel.indexOf(Util.SEGMENT_TIER_SEPARATOR) + 1);

        segmentSelectionPanel.getUtteranceView().addUtterancePane(null, temporalPartitionLabel, segmentTierLabel);
        segmentSelectionPanel.getParentDialog().dispose();
	}
}
